from dataclasses import dataclass, field, asdict
from typing import List, Optional, Union, Dict
from tqdm import tqdm
import pandas as pd

from bioel.logger import setup_logger
from bioel.utils.bigbio_utils import dataset_unique_tax_ids

import obonet
import csv
from bioel.utils.obo_utils import _obo_extract_definition, _obo_extract_synonyms
from bioel.utils.umls_utils import UmlsMappings
import warnings
import ujson
import json
from bioel.utils.file_cache import get_path

logger = setup_logger()


@dataclass
class BiomedicalEntity:
    """
    Class for keeping track of all relevant fields in an ontology
    """

    cui: str
    name: str
    types: List[str]
    aliases: List[str] = field(default_factory=list)
    definition: Optional[str] = None
    equivalant_cuis: Optional[List[str]] = None
    taxonomy: Optional[str] = None
    metadata: Optional[dict] = None


@dataclass
class BiomedicalOntology:
    name: str
    types: List[str] = field(default_factory=list)
    entities: Dict[str, BiomedicalEntity] = field(
        default_factory=dict
    )  # Dict mapping CUI: BiomedicalEntity
    abbrev: Optional[str] = None  # Abbreviated name of ontology if different than name
    metadata: Optional[dict] = None
    dataset: Optional[str] = (
        None  # Name of the dataset to work with (only for Entrez ontology)
    )

    def get_canonical_name(self):
        """
        Get name of entities in the ontology
        """
        canonical_names = {cui: entity.name for cui, entity in self.entities.items()}
        return canonical_names

    def get_aliases(self):
        """
        Get aliases of entities in the ontology
        """
        aliases_dict = {cui: entity.aliases for cui, entity in self.entities.items()}
        return aliases_dict

    def get_definition(self):
        """
        Get definition of entities in the ontology
        """
        definitions_dict = {
            entity.cui: entity.definition
            for entity in self.entities.values()
            if entity.definition is not None
        }
        return definitions_dict

    def get_types(self):
        """
        Get types of entities in the ontology
        """
        types_dict = {cui: entity.types for cui, entity in self.entities.items()}
        return types_dict

    @classmethod
    def load_obo(
        cls, filepath, name=None, prefix_to_keep=None, entity_type=None, abbrev=None
    ):
        """
        Read an ontology in .obo format

        Parameters:
        ----------------------
            filepath: str (optional)
                Path to .obo formatted ontology.  Can be a URL.
            prefix_to_keep: str (optional)
        """

        entities = {}
        if entity_type:
            types = [entity_type]
        else:
            types = []

        logger.info(f"Reading OBO ontology from {filepath}")
        ontology = obonet.read_obo(filepath)

        data_keys = set([])
        for curie, data in tqdm(ontology.nodes(data=True)):
            # Exclude CUIs that are from cross-referenced ontologies
            data_keys.update(set(data.keys()))

            if prefix_to_keep is not None:
                if not curie.startswith(prefix_to_keep):
                    continue

            # Extract name and synonyms
            if "name" not in data:
                synonyms = _obo_extract_synonyms(data)
            else:
                synonyms = [data["name"]] + _obo_extract_synonyms(data)

            # Include deprecated CUIs as alternative ids
            alt_cuis = None
            if "alt_id" in data and len(data["alt_id"]) > 0:
                if prefix_to_keep:
                    alt_cuis = [
                        x for x in data["alt_id"] if x.startswith(prefix_to_keep)
                    ]
                else:
                    alt_cuis = data["alt_id"]

            # Skip entities with no name/aliases
            if len(synonyms) == 0:
                logger.warning(
                    f"Data entry for CUI {curie} has no listed names/aliases.  Skipping."
                )
                continue
            ent_name = synonyms[0]
            if len(synonyms) > 1:
                other_synonyms = synonyms[1:]
            else:
                other_synonyms = []

            # Get definition if it exists
            definition = _obo_extract_definition(data)

            ent = BiomedicalEntity(
                cui=curie,
                name=ent_name,
                aliases=other_synonyms,
                types=types,
                definition=definition,
                equivalant_cuis=alt_cuis,
            )
            if curie in entities:
                logger.warning(f"Duplicate CUI {curie} found in ontology.  Skipping.")
                continue

            entities[curie] = ent

        if not name:
            if filepath.startswith("http"):
                parsed_name = filepath.split("/")[-1].split(".")[0]
                logger.warning(
                    f"No ontology name provided.  Using name from provided URL: {parsed_name}"
                )
                name = parsed_name

        return cls(entities=entities, types=types, name=name, abbrev=abbrev)

    @classmethod
    def load_medic(cls, filepath, name=None, abbrev=None, api_key=""):
        """
        Read medic ontology

        Parameters:
        ----------------------
            filepath: str (Pointing to the medic directory)
            name: str (optional)
            abbrev: str (optional)
            api_key: str (optional)
        """
        entities = {}
        types = []

        logger.info(f"Reading entrez from {filepath}")
        col_names = "Disease_Name   Disease_ID       Alt_Disease_IDs   Definition      Parent_IDs       Tree_Numbers     Parent_Tree_Numbers       Synonyms        Slim_Mappings".split()

        # Open the TSV file
        medic = pd.read_csv(
            filepath,
            comment="#",
            delimiter="\t",
            na_filter=False,
            names=col_names,
            usecols=[
                "Disease_Name",
                "Disease_ID",
                "Alt_Disease_IDs",
                "Definition",
                "Parent_IDs",
                "Tree_Numbers",
                "Parent_Tree_Numbers",
                "Synonyms",
                "Slim_Mappings",
            ],
        )
        for index, row in medic.iterrows():
            equivalant_cuis = [row["Disease_ID"]]
            alt_ids = (
                row["Alt_Disease_IDs"].split("|") if row["Alt_Disease_IDs"] else []
            )
            for alt_id in alt_ids:
                if alt_id not in equivalant_cuis and alt_id[:2] != "DO":
                    equivalant_cuis.append(alt_id)

            entity = BiomedicalEntity(
                cui=row["Disease_ID"],
                name=row["Disease_Name"],
                types="Disease",
                aliases=row["Synonyms"],
                definition=row["Definition"],
                equivalant_cuis=equivalant_cuis,
            )

            if row["Disease_ID"] in entities:
                logger.warning(
                    f"Duplicate CUI {row['Disease_ID']} found in ontology.  Skipping."
                )
                continue

            entities[row["Disease_ID"]] = entity

            types.append("Disease")
        return cls(entities=entities, types=types, name=name, abbrev=abbrev)

    @classmethod
    def load_entrez(cls, filepath, dataset, name=None, abbrev=None, api_key=""):
        """
        Read medic ontology

        Parameters:
        ----------------------
            filepath: str (Pointing to the medic directory)
            name: str (optional)
            abbrev: str (optional)
            api_key: str (optional)
        """

        entities = {}
        types = []

        logger.info(f"Reading entrez from {filepath}")

        # Open the TSV file
        entrez = pd.read_csv(
            filepath,
            delimiter="\t",
            usecols=[
                "#tax_id",
                "GeneID",
                "Symbol",
                "Synonyms",
                "Symbol_from_nomenclature_authority",
                "Full_name_from_nomenclature_authority",
                "Other_designations",
                "type_of_gene",
                "description",
                "dbXrefs",
            ],
            na_filter=False,
            low_memory=False,
        ).rename(
            {
                "Symbol_from_nomenclature_authority": "official_symbol",
                "Full_name_from_nomenclature_authority": "official_name",
                "#tax_id": "tax_id",
            },
            axis=1,
        )
        entrez.columns = [x.lower() for x in entrez.columns]

        unique_tax_ids = dataset_unique_tax_ids(dataset, entrez)

        geneid_mask = (
            (entrez.tax_id.isin(unique_tax_ids))
            & (~entrez.type_of_gene.isin(["unknown", "tRNA", "biological-region"]))
            & (entrez.description != "hypothetical protein")
            & (~entrez.official_name.map(lambda x: x.lower().startswith("predicted")))
        )
        entrez = entrez[geneid_mask]

        entrez.replace("-", "", inplace=True)
        entrez["geneid"] = entrez["geneid"].map(lambda x: f"NCBIGene:{x}")
        entrez["metadata"] = entrez[
            [
                "official_symbol",
                "official_name",
            ]
        ].progress_apply(
            lambda x: ";".join(list(set([i for i in x if i.strip() != "-"]))), axis=1
        )

        for index, row in entrez.iterrows():

            entity = BiomedicalEntity(
                cui=row["geneid"],
                name=row["symbol"],
                types=row["type_of_gene"],
                aliases=row["synonyms"],
                definition=row["description"],
                taxonomy=row["tax_id"],
                metadata=row["metadata"],
            )

            if row["geneid"] in entities:
                logger.warning(
                    f"Duplicate CUI {row['geneid']} found in ontology.  Skipping."
                )
                continue

            entities[row["geneid"]] = entity

            types.append(row["type_of_gene"])

        return cls(entities=entities, types=types, name=name, abbrev=abbrev)
    
    @classmethod
    def load_umls(
        cls, filepath, path_st21pv_cui=None, name=None, abbrev=None, api_key=""
    ):
        """
        Read an ontology from the UMLS Directory

        Parameters:
        ----------------------
            filepath: str (Pointing to the UMLS directory)
            path_st21pv_cui : str (optional : Path to st21pv cuis subset for umls)
            name: str (optional)
            abbrev: str (optional)
            api_key: str (optional)
        """

        entities = {}
        types = []

        logger.info(f"Reading UMLS from {filepath}")
        umls = UmlsMappings(umls_dir=filepath, umls_api_key=api_key)

        # Get the Canonial Names
        lowercase = False
        umls_to_name = umls.get_canonical_name(
            ontologies_to_include="all",
            use_umls_curies=True,
            lowercase=lowercase,
        )

        # Group by the canonical names to group the alias and types
        all_umls_df = (
            umls.umls.query('lang == "ENG"')
            .groupby("cui")
            .agg(
                {
                    "alias": lambda x: list(set(x)),
                    "tui": "first",
                    "group": "first",
                    "def": "first",
                }
            )
            .reset_index()
        )

        if path_st21pv_cui:
            with open(path_st21pv_cui, "r") as file:
                umls_cuis_st21pv = ujson.load(file)
            all_umls_df = all_umls_df[
                all_umls_df["cui"].isin(umls_cuis_st21pv)
            ].reset_index()

        all_umls_df["name"] = all_umls_df.cui.map(umls_to_name)
        all_umls_df["alias"] = all_umls_df[["name", "alias"]].apply(
            lambda x: list(set(x[1]) - set([x[0]])), axis=1
        )
        all_umls_df["cui"] = all_umls_df["cui"].map(lambda x: "UMLS:" + x)
        all_umls_df["has_definition"] = all_umls_df["def"].map(lambda x: x is not None)
        all_umls_df["num_aliases"] = all_umls_df["alias"].map(lambda x: len(x))

        print("Number of entities :", all_umls_df.shape[0])

        for index, row in tqdm(all_umls_df.iterrows(), desc="Loading UMLS Ontology"):
            entity = BiomedicalEntity(
                cui=row["cui"],
                name=row["name"],
                types=row["tui"],
                aliases=row["alias"],
                definition=row["def"],
                metadata={
                    "group": row["group"],
                },
            )
            if row["cui"] in entities:
                logger.warning(
                    f"Duplicate CUI {row['cui']} found in ontology.  Skipping."
                )
                continue

            entities[row["cui"]] = entity
            types.append(row["tui"])

        return cls(entities=entities, types=types, name=name, abbrev=abbrev)

    @classmethod
    def load_entrez(cls, filepath, dataset, name=None, abbrev=None, api_key=""):
        """
        Read medic ontology

        Parameters:
        ----------------------
            filepath: str (Pointing to the medic directory)
            name: str (optional)
            abbrev: str (optional)
            api_key: str (optional)
        """

        entities = {}
        types = []

        logger.info(f"Reading entrez from {filepath}")

        # Open the TSV file
        entrez = pd.read_csv(
            filepath,
            delimiter="\t",
            usecols=[
                "#tax_id",
                "GeneID",
                "Symbol",
                "Synonyms",
                "Symbol_from_nomenclature_authority",
                "Full_name_from_nomenclature_authority",
                "Other_designations",
                "type_of_gene",
                "description",
                "dbXrefs",
            ],
            na_filter=False,
            low_memory=False,
        ).rename(
            {
                "Symbol_from_nomenclature_authority": "official_symbol",
                "Full_name_from_nomenclature_authority": "official_name",
                "#tax_id": "tax_id",
            },
            axis=1,
        )
        entrez.columns = [x.lower() for x in entrez.columns]

        unique_tax_ids = dataset_unique_tax_ids(dataset, entrez)

        geneid_mask = (
            (entrez.tax_id.isin(unique_tax_ids))
            & (~entrez.type_of_gene.isin(["unknown", "tRNA", "biological-region"]))
            & (entrez.description != "hypothetical protein")
            & (~entrez.official_name.map(lambda x: x.lower().startswith("predicted")))
        )
        entrez = entrez[geneid_mask]

        entrez.replace("-", "", inplace=True)
        entrez["geneid"] = entrez["geneid"].map(lambda x: f"NCBIGene:{x}")
        entrez["metadata"] = entrez[
            [
                "official_symbol",
                "official_name",
            ]
        ].progress_apply(
            lambda x: ";".join(list(set([i for i in x if i.strip() != "-"]))), axis=1
        )

        for index, row in entrez.iterrows():

            entity = BiomedicalEntity(
                cui=row["geneid"],
                name=row["symbol"],
                types=row["type_of_gene"],
                aliases=row["synonyms"],
                definition=row["description"],
                taxonomy=row["tax_id"],
                metadata=row["metadata"],
            )

            if row["geneid"] in entities:
                logger.warning(
                    f"Duplicate CUI {row['geneid']} found in ontology.  Skipping."
                )
                continue

            entities[row["geneid"]] = entity

            types.append(row["type_of_gene"])

        return cls(entities=entities, types=types, name=name, abbrev=abbrev)

    @classmethod
    def load_mesh(cls, filepath, name=None, abbrev=None, api_key=""):
        """
        Read an ontology from the MESH Directory

        Parameters:
        ----------------------
            filepath: str (Pointing to the UMLS directory)
            name: str (optional)
            abbrev: str (optional)
            api_key: str (optional)
        """
        entities = {}
        types = []

        logger.info(f"Reading MESH from : {filepath}")
        umls = UmlsMappings(umls_dir=filepath, umls_api_key=api_key)

        # Get the Canonial Names
        lowercase = False
        mesh_to_name = umls.get_canonical_name(
            ontologies_to_include=["MSH"],
            use_umls_curies=False,
            mapping_cols={"MSH": "sdui"},
            prefixes={"MSH": "MESH"},
            lowercase=lowercase,
        )

        mesh_to_alias = umls.get_aliases(
            ontologies_to_include=["MSH"],
            use_umls_curies=False,
            mapping_cols={"MSH": "sdui"},
            prefixes={"MSH": "MESH"},
            lowercase=lowercase,
        )

        mesh_cui2definition = umls.get_definition(
            ontologies_to_include=["MSH"],
            use_umls_curies=False,
            mapping_cols={"MSH": "sdui"},
            prefixes={"MSH": "MESH"},
            lowercase=lowercase,
        )

        mesh_to_types, mesh_to_groups = umls.get_types_and_groups(
            ontologies_to_include=["MSH"],
            use_umls_curies=False,
            mapping_cols={"MSH": "sdui"},
            prefixes={"MSH": "MESH"},
            lowercase=lowercase,
        )
        i = 0
        for cui, _name in tqdm(mesh_to_name.items()):
            # ent_type = mesh_to_types[cui]
            ent_type = mesh_to_groups[cui][0]
            # if i < 1:
            #     print(f"{mesh_to_groups[cui][0]=}")
            #     print(f"{mesh_to_types[cui]=}")
            other_aliases = [x for x in mesh_to_alias[cui] if x != _name]
            joined_aliases = " ; ".join(other_aliases)
            entity = BiomedicalEntity(
                cui=cui,
                name=_name,
                types=ent_type,
                aliases=joined_aliases,
                definition=(
                    mesh_cui2definition[cui] if cui in mesh_cui2definition else None
                ),
                metadata={
                    "group": mesh_to_groups[cui],
                },
            )

            if cui in entities:
                logger.warning(f"Duplicate CUI {cui} found in ontology.  Skipping.")
                continue

            entities[cui] = entity
            types.append(ent_type)

        return cls(entities=entities, types=types, name=name, abbrev=abbrev)
        
    @classmethod
    def load_json(cls, filepath, name=None):
        '''
        file_path: str, required.
            The file path to the json/jsonl representation of the KB to load.
        name: str, optional.
            The ontology name to load.
        '''
        if filepath is None:
            raise ValueError(
                "provide a valid path"
            )
        if filepath.endswith("jsonl"):
            raw = (json.loads(line) for line in open(get_path(filepath)))
        else:
            raw = json.load(open(get_path(filepath)))

        logger.info(f"Reading the given Json ontology from {filepath}")
        
        index = 0
        types = []
        entities = {}
        metadata ={}
        if name is None:
            name = "json_onto"
        for concept in raw:
            types.append(concept["types"])
            metadata[concept["concept_id"]] = index
            index += 1
            if concept["concept_id"] in entities:
                logger.warning(f"Duplicate CUI {concept['concept_id']} found in ontology.  Skipping.")
                continue
            elif 'definition' in concept:
                entities[concept["concept_id"]] = BiomedicalEntity(concept["concept_id"], concept["canonical_name"], concept["types"], concept["aliases"], concept["definition"])
            else:
                entities[concept["concept_id"]] = BiomedicalEntity(concept["concept_id"], concept["canonical_name"], concept["types"], concept["aliases"])
                
        return cls(entities = entities, types = types, metadata = metadata, name = name)

    def load_ncbi_taxon(self):
        pass

    def load_csv(self):
        pass

    def load_json(self):
        pass


@dataclass
class CompositeOntology:
    ontologies: dict  # Dict of each ontology used

    def load_from_config(self, config=None, config_path=None):
        pass



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import torch
import numpy as np
from scipy.sparse import coo_matrix
from scipy.sparse.csgraph import connected_components
from bioel.models.arboel.biencoder.model.special_partition.special_partition import (
    cluster_linking_partition,
)

from collections import defaultdict
import bioel.models.arboel.biencoder.data.data_process as data_process
from bioel.models.arboel.biencoder.model.biencoder import BiEncoderRanker

from IPython import embed


def get_query_nn(
    knn, embeds, index, q_embed, searchK=None, gold_idxs=None, type_idx_mapping=None
):
    """
    Retrieves the top k neighbours of the query embedding (We know it's for mentions because of knn)

    Parameters
    ----------
    knn : int
        the number of nearest-neighbours to return
    embeds : ndarray of dim (number_mentions, H: embedding_size)
        matrix of embeddings
    index : faiss
        faiss index of the embeddings
    q_embed : ndarray of dim (1, H)
        2-D array containing the query embedding
    searchK: int
        optional parameter, the exact number of nearest-neighbours to retrieve and score
        (Used for approximate search)
    gold_idxs : array
        optional parameter, list of golden cui indexes (= the correct mentions for the query).
    type_idx_mapping : array
        optional parameter, list mapping type-specific indexes to the indexes of the full dictionary

    Returns
    -------
    nn_idxs : array
        indices of the nearest neighbour, sorted in descending order of scores
    scores : array
        similarity scores of the nearest neighbors, sorted in descending order
    """
    # To accomodate the approximate-nature of the knn procedure, retrieve more samples and then filter down
    k = searchK if searchK is not None else max(16, 2 * knn)

    # Find k nearest neighbours of the query
    _, nn_idxs = index.search(q_embed, k)
    nn_idxs = nn_idxs.astype(np.int64).flatten()
    if type_idx_mapping is not None:
        nn_idxs = type_idx_mapping[nn_idxs]
    nn_embeds = torch.tensor(embeds[nn_idxs]).cuda()

    # Compute query-candidate similarity scores
    scores = torch.flatten(torch.mm(torch.tensor(q_embed).cuda(), nn_embeds.T)).cpu()

    # Sort the candidates by descending order of scores
    nn_idxs, scores = zip(*sorted(zip(nn_idxs, scores), key=lambda x: -x[1]))

    if gold_idxs is not None:
        # Calculate the knn index at which the gold cui is found (-1 if not found)
        # Only find one even though several can exist
        for topk, i in enumerate(nn_idxs):
            if i in gold_idxs:
                break
            topk = -1
        # Return only the top k neighbours, and the recall index
        return (
            np.array(nn_idxs[:knn], dtype=np.int64),
            np.array(scores[:knn]),
            topk,
        )

    # Return only the top k neighbours
    return np.array(nn_idxs[:knn], dtype=np.int64), np.array(scores[:knn])


def partition_graph(
    graph, n_entities, directed, return_clusters=False, exclude=set(), threshold=None
):
    """
    Convert the initial graph constructed in train_biencoder_mst.evaluate(.) into the final graph resulting from the inference procedure

    Parameters
    ----------
    graph : dict
        object containing rows, cols, data, and shape of the entity-mention joint graph
    n_entities : int
        number of entities in the dictionary
    directed : bool
        whether the graph construction should be directed or undirected
    return_clusters : bool
        flag to indicate if clusters need to be returned from the partition
    exclude : set
        vertices (both rows and columns) whose edges should be dropped in the graph
    threshold : float
        similarity value below which edges should be dropped


    Returns
    -------
    partitioned_graph : coo_matrix
        partitioned graph with each mention connected to only one entity
    clusters : dict
        (optional) contains arrays of connected component indices of the graph
    """
    rows, cols, data, shape = (
        graph["rows"],
        graph["cols"],
        graph["data"],
        graph["shape"],
    )

    rows, cols, data = cluster_linking_partition(
        rows, cols, data, n_entities, directed, exclude=exclude, threshold=threshold
    )
    # Construct the partitioned graph
    partitioned_graph = coo_matrix((data, (rows, cols)), shape=shape)

    if return_clusters:
        # Get an array of the graph with each index marked with the component label that it is connected to
        _, cc_labels = connected_components(
            csgraph=partitioned_graph, directed=directed, return_labels=True
        )
        # Store clusters of indices marked with labels with at least 2 connected components
        unique_cc_labels, cc_sizes = np.unique(cc_labels, return_counts=True)
        filtered_labels = unique_cc_labels[cc_sizes >= 2]
        clusters = defaultdict(list)
        for i, cc_label in enumerate(cc_labels):
            if cc_label in filtered_labels:
                clusters[cc_label].append(i)
        return partitioned_graph, clusters

    return partitioned_graph


def analyzeClusters(clusters, dictionary, queries, knn, n_train_mentions=0):
    """
    Evaluates the accuracy of entity-mention clustering

    Parameters
    ----------
    clusters : dict
        contains arrays of connected component indices of a graph
    dictionary : ndarray
        entity dictionary to evaluate
    queries : ndarray
        mention queries to evaluate
    knn : int
        the number of nearest-neighbour mention candidates considered

    Returns
    -------
    results : dict
        Contains n_entities, n_mentions, knn_mentions, accuracy, failure[], success[]
    """
    n_entities = len(dictionary)
    n_mentions = len(queries)

    results = {
        "n_entities": n_entities,
        "n_mentions": n_mentions,
        "knn_mentions": knn,
        "accuracy": 0,
        "failure": [],
        "success": [],
    }
    (
        _debug_n_mens_evaluated,
        _debug_clusters_wo_entities,
        _debug_clusters_w_mult_entities,
    ) = (0, 0, 0)

    for cluster in clusters.values():
        "Evaluate entity"
        # The lowest value in the cluster should always be the entity
        pred_entity_idx = cluster[0]
        # Track the graph index of the entity in the cluster
        pred_entity_idxs = [pred_entity_idx]
        if pred_entity_idx >= n_entities:
            # If the first element is a mention, then the cluster does not have an entity
            _debug_clusters_wo_entities += 1
            continue  # Skip to the next cluster
        pred_entity = dictionary[pred_entity_idx]
        pred_entity_cuis = [
            str(pred_entity["cui"])
        ]  # Retrieve the cui of the entity indice
        _debug_tracked_mult_entities = False
        for i in range(1, len(cluster)):
            men_idx = cluster[i] - n_entities
            if men_idx < 0:
                # If elements after the first are entities, then the cluster has multiple entities
                if not _debug_tracked_mult_entities:
                    _debug_clusters_w_mult_entities += 1
                    _debug_tracked_mult_entities = True
                # Track the graph indices of each entity in the cluster
                pred_entity_idxs.append(cluster[i])
                # Predict based on all entities in the cluster
                pred_entity_cuis += list(
                    set([dictionary[cluster[i]]["cui"]]) - set(pred_entity_cuis)
                )
                continue
            men_idx -= n_train_mentions  # number of training mentions
            if men_idx < 0:
                # Query is from train set
                continue
            " For each mentions in the cluster, we compare its mention_gold_cui to the predicted_cui"
            _debug_n_mens_evaluated += 1
            men_query = queries[men_idx]
            men_golden_cuis = list(
                map(str, men_query["label_cuis"])
            )  # One mention can can be associated with several CUIs in certain contexts
            report_obj = {
                "mention_id": men_query["mention_id"],
                "mention_name": men_query["mention_name"],
                "mention_gold_cui": "|".join(men_golden_cuis),
                "mention_gold_cui_name": "|".join(
                    [
                        dictionary[i]["title"]
                        for i in men_query["label_idxs"][: men_query["n_labels"]]
                    ]
                ),
                "predicted_name": "|".join(
                    [d["title"] for d in [dictionary[i] for i in pred_entity_idxs]]
                ),
                "predicted_cui": "|".join(pred_entity_cuis),
            }
            # Correct prediction = if there is at least one common element between the predicted entity CUIs (pred_entity_cuis) and the gold standard entity CUIs (men_golden_cuis)
            if not set(pred_entity_cuis).isdisjoint(men_golden_cuis):
                results["accuracy"] += 1
                results["success"].append(report_obj)
            # Incorrect prediction
            else:
                results["failure"].append(report_obj)
    results["accuracy"] = (
        f"{results['accuracy'] / float(_debug_n_mens_evaluated) * 100} %"
    )
    print(f"Accuracy = {results['accuracy']}")

    # Run sanity checks
    assert _debug_clusters_wo_entities == 0
    assert _debug_clusters_w_mult_entities == 0
    assert n_mentions == _debug_n_mens_evaluated

    return results


def filter_by_context_doc_id(mention_idxs, doc_id, doc_id_list, return_numpy=False):
    """
    Description
    -----------
    Filters and returns only mention indices that belong to a specific document identified by the doc_id.
    Ensures that the analysis are constrained within the context of that particular document.

    Parameters
    ----------
    - mention_idxs : ndarray(int) of dim = (number of mentions)
    Represents the indices of mentions
    - doc_id : int
    Indice of the target document
    - doc_id_list : ndarray(int) of dim = (number of mentions)
    Array of integers, where each element is a document ID associated with the corresponding mention in mention_idxs.
    The length of doc_id_list should match the total number of mentions referenced in mention_idxs.
    - return_numpy : bool
    A flag indicating whether to return the filtered list of mention indices as a NumPy array.
    If True, the function returns a NumPy array; otherwise, it returns a list
    -------
    Outputs:
    - mask : ndarray(bool) of dim = (number of mentions)
    Mask indicating where each mention's document ID (from "doc_id_list") matches the target "doc_id"
    - mention_idxs :
    Only contains mention indices that belong to the target document (=doc_id).
    """
    mask = [doc_id_list[i] == doc_id for i in mention_idxs]
    if isinstance(mention_idxs, list):
        mention_idxs = np.array(mention_idxs)
    mention_idxs = mention_idxs[mask]
    if not return_numpy:
        mention_idxs = list(mention_idxs)
    return mention_idxs, mask



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import logging
from tqdm import tqdm
import random
from torch.cuda.amp import autocast
from pytorch_metric_learning import miners, losses, distances
LOGGER = logging.getLogger(__name__)


class Sap_Metric_Learning(nn.Module):
    def __init__(self, encoder, learning_rate, weight_decay, use_cuda, pairwise, 
            loss, use_miner=True, miner_margin=0.2, type_of_triplets="all", agg_mode="cls"):

        LOGGER.info("Sap_Metric_Learning! learning_rate={} weight_decay={} use_cuda={} loss={} use_miner={} miner_margin={} type_of_triplets={} agg_mode={}".format(
            learning_rate,weight_decay,use_cuda,loss,use_miner,miner_margin,type_of_triplets,agg_mode
        ))
        super(Sap_Metric_Learning, self).__init__()
        self.encoder = encoder
        self.pairwise = pairwise
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.use_cuda = use_cuda
        self.loss = loss
        self.use_miner = use_miner
        self.miner_margin = miner_margin
        self.agg_mode = agg_mode
        self.optimizer = optim.AdamW([{'params': self.encoder.parameters()},], 
            lr=self.learning_rate, weight_decay=self.weight_decay
        )
        
        if self.use_miner:
            self.miner = miners.TripletMarginMiner(margin=miner_margin, type_of_triplets=type_of_triplets)
        else:self.miner = None

        if self.loss == "ms_loss":
            self.loss = losses.MultiSimilarityLoss(alpha=1, beta=60, base=0.5) # 1,2,3; 40,50,60
        elif self.loss == "circle_loss":
            self.loss = losses.CircleLoss()
        elif self.loss == "triplet_loss":
            self.loss = losses.TripletMarginLoss()
        elif self.loss == "infoNCE":
            self.loss = losses.NTXentLoss(temperature=0.07) # The MoCo paper uses 0.07, while SimCLR uses 0.5.
        elif self.loss == "lifted_structure_loss":
            self.loss = losses.LiftedStructureLoss()
        elif self.loss == "nca_loss":
            self.loss = losses.NCALoss()


        print ("miner:", self.miner)
        print ("loss:", self.loss)
    
    @autocast() 
    def forward(self, query_toks1, query_toks2, labels):
        """
        query : (N, h), candidates : (N, topk, h)

        output : (N, topk)
        """
        
        last_hidden_state1 = self.encoder(**query_toks1, return_dict=True).last_hidden_state
        last_hidden_state2 = self.encoder(**query_toks2, return_dict=True).last_hidden_state
        if self.agg_mode=="cls":
            query_embed1 = last_hidden_state1[:,0]  # query : [batch_size, hidden]
            query_embed2 = last_hidden_state2[:,0]  # query : [batch_size, hidden]
        elif self.agg_mode == "mean_all_tok":
            query_embed1 = last_hidden_state1.mean(1)  # query : [batch_size, hidden]
            query_embed2 = last_hidden_state2.mean(1)  # query : [batch_size, hidden]
        elif self.agg_mode == "mean":
            query_embed1 = (last_hidden_state1 * query_toks1['attention_mask'].unsqueeze(-1)).sum(1) / query_toks1['attention_mask'].sum(-1).unsqueeze(-1)
            query_embed2 = (last_hidden_state2 * query_toks2['attention_mask'].unsqueeze(-1)).sum(1) / query_toks2['attention_mask'].sum(-1).unsqueeze(-1)
        else:
            raise NotImplementedError()
        query_embed = torch.cat([query_embed1, query_embed2], dim=0)
        
        labels = torch.cat([labels, labels], dim=0)
        
        if self.use_miner:
            hard_pairs = self.miner(query_embed, labels)
            return self.loss(query_embed, labels, hard_pairs) 
        else:
            return self.loss(query_embed, labels) 


    def reshape_candidates_for_encoder(self, candidates):
        """
        reshape candidates for encoder input shape
        [batch_size, topk, max_length] => [batch_size*topk, max_length]
        """
        _, _, max_length = candidates.shape
        candidates = candidates.contiguous().view(-1, max_length)
        return candidates

    def get_loss(self, outputs, targets):
        if self.use_cuda:
            targets = targets.cuda()
        loss, in_topk = self.criterion(outputs, targets)
        return loss, in_topk

    def get_embeddings(self, mentions, batch_size=1024):
        """
        Compute all embeddings from mention tokens.
        """
        embedding_table = []
        with torch.no_grad():
            for start in tqdm(range(0, len(mentions), batch_size)):
                end = min(start + batch_size, len(mentions))
                batch = mentions[start:end]
                batch_embedding = self.vectorizer(batch)
                batch_embedding = batch_embedding.cpu()
                embedding_table.append(batch_embedding)
        embedding_table = torch.cat(embedding_table, dim=0)
        return embedding_table






from bioel.models.arboel.crossencoder.data.CrossEncoderLightningDataModule import (
    CrossEncoderDataModule,
    prepare_data,
)
from bioel.models.arboel.crossencoder.model.CrossEncoderLightningModule import (
    LitCrossEncoder,
)
from datetime import datetime, timedelta
from lightning.pytorch.strategies.ddp import DDPStrategy
from lightning.pytorch.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import WandbLogger
import lightning as L
from bioel.models.arboel.biencoder.model.common.params import BlinkParser
import os
import torch


def train_model(params, model):
    print("Current seed:", params["seed"])

    prepare_data(params)

    data_module = CrossEncoderDataModule(params=params)

    current_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    model_checkpoint = ModelCheckpoint(
        monitor="Accuracy",  # Metric to monitor
        dirpath=params["output_path"],  # Directory to save the model
        filename=f"crossencoder_{current_time}-{{epoch}}-{{Accuracy:.2f}}",  # Saves the model with epoch and val_loss in the filename
        save_top_k=1,  # Number of best models to save; -1 means save all of them
        mode="max",  # 'max' means the highest max_acc will be considered as the best model
        verbose=True,  # Logs a message whenever a model checkpoint is saved
    )

    wandb_logger = WandbLogger(project=params["experiment"])

    trainer = L.Trainer(
        limit_train_batches=params.get("limit_train_batches", 1.0),
        limit_val_batches=params.get("limit_val_batches", 20),
        # num_sanity_val_steps=0,
        max_epochs=params["num_train_epochs"],
        devices=params["devices"],
        accelerator="gpu" if params["devices"] else "cpu",
        strategy="ddp_find_unused_parameters_true",
        enable_progress_bar=True,
        callbacks=[model_checkpoint],
        accumulate_grad_batches=params["gradient_accumulation_steps"],
        precision="16-mixed",
        check_val_every_n_epoch=1,
        logger=wandb_logger,
    )

    trainer.fit(model, datamodule=data_module)



from tqdm import tqdm
import itertools
import random
from typing import List, Optional
import torch
import logging
import os
import glob
import numpy as np
import json
import pandas as pd
import ujson
import warnings
from itertools import combinations
from collections import defaultdict

from torch.utils.data import Dataset, DataLoader
from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    dataset_to_df,
    DATASET_NAMES,
    CUIS_TO_EXCLUDE,
    CUIS_TO_REMAP,
    resolve_abbreviation,
    add_deabbreviations,
)

from transformers import AutoTokenizer

from bioel.ontology import BiomedicalOntology
from bioel.logger import setup_logger

logger = setup_logger()


"""
Generate Pretraining Data for SAPBERT from the UMLS Ontology
"""
def generate_positive_pairs(alias_mapping: dict, max_pairs_per_cui: int = 50) -> List[str]:
    """
    Generate pretraining data from UMLS for the SAPBERT Contrastive Learning.
    """
    positive_pairs = []
    for cui, entity in tqdm(alias_mapping.items(), desc = "Generating pretraining data for SAPBERT"):
        if isinstance(entity, list):
            pairs = list(combinations(entity, r = 2))
        else:
            pairs = list(combinations(entity.aliases, r = 2))
        if len(pairs) > max_pairs_per_cui: # If > 50 pairs, sample 50 pairs
            pairs = random.sample(pairs, k=max_pairs_per_cui)
        
        for p in pairs:
            if p[0] and p[1]:
                line = str(cui) + "||" + p[0].lower() + "||" + p[1].lower()
                
                positive_pairs.append(line)
    return positive_pairs

class DictionaryDataset:
    """
    A class used to load dictionary data
    """

    def __init__(self, dictionary_path):
        """
        Parameters
        ----------
        dictionary_path : str
            The path of the dictionary
        draft : bool
            use only small subset
        """
        logger.info("DictionaryDataset! dictionary_path={}".format(dictionary_path))
        self.data = self.load_data(dictionary_path)

    def load_data(self, dictionary_path):
        name_cui_map = {}
        data_dict = defaultdict(list)
        with open(dictionary_path, mode="r", encoding="utf-8") as f:
            lines = f.readlines()
            for line in tqdm(lines):
                line = line.strip()
                if not line or len(line.split("||")) != 2:
                    continue
                if line == "":
                    continue
                cui, name = line.split("||")
                name = name.lower()
                if cui.lower() == "cui-less":
                    continue

                data_dict[name].append(cui)
                # data.append((name,cui))

        # LOGGER.info("concerting loaded dictionary data to numpy array...")
        # data = np.array(data)
        data = [(name, "|".join(cuis)) for name, cuis in data_dict.items()]
        return data

class SapBertDictionaryDataset(Dataset):
    def __init__(self, dictionary_path: str):
        logger.info("Loading from the Alias Mapping of the Dataset")
        self.data = self.read_examples(dictionary_path)

    
    def read_examples(self, dictionary_path):
        """
        Read examples from the file
        """
        with open(dictionary_path, "r") as f:
            lines = f.read().split("\n")
        
        # Construt the UMLS Dict for finetuning the model
        umls_dict = {}
        for line in tqdm(lines, desc="Reading examples from the file"):
            if not line or len(line.split("||")) != 2:
                continue
            cui, name = line.split("||")
            name = name.lower()
            if cui in umls_dict:
                umls_dict[cui].add(name)
            else:
                umls_dict[cui] = set([name])
        
        for key, value in umls_dict.items():
            umls_dict[key] = list(value)
        
        return umls_dict
        

        
    
class SapBertBigBioDataset(Dataset):
    def __init__(
        self,
        dataset_name: str = "medmentions_st21pv",
        splits_to_include: List[str] = ["train"],
        resolve_abbreviations: bool = True,
        path_to_abbreviation_dict: Optional[str] = None,
    ):
        """
        Load initial BigBio dataset
        """
        # Pull in data
        self.data = load_bigbio_dataset(dataset_name)
        self.splits_to_include = splits_to_include

        # Resolve abbreviations if desired
        self.resolve_abbreviations = resolve_abbreviations
        if self.resolve_abbreviations:
            self.data = add_deabbreviations(self.data, path_to_abbreviation_dict)
        #     self.abbreviations = ujson.load(open(path_to_abbreviation_dict, "r"))

        self.cuis_to_exclude = CUIS_TO_EXCLUDE[dataset_name]
        self.cuis_to_remap = CUIS_TO_REMAP[dataset_name]

        # Put examples into list for retrieval
        self._data_to_flat_instances()



    def _data_to_flat_instances(self):
        """
        Convert dataset into flat set of positive Examples to use with dataloader
        """
        df = dataset_to_df(
            self.data,
            self.splits_to_include,
            entity_remapping_dict=self.cuis_to_remap,
            cuis_to_exclude=self.cuis_to_exclude,
        )
        # if self.resolve_abbreviations:
        #     df["text"] = df[["document_id", "text"]].apply(
        #         lambda x: resolve_abbreviation(x[0], x[1], self.abbreviations), axis=1
        #     )
        self.flat_instances = df.to_dict(orient="records")

    

    def __len__(self):
        return len(self.flat_instances)

    def __getitem__(self, idx):
        return self.flat_instances[idx]


# def sapbert_collate_fn(batch):
#     mentions = [x["text"] for x in batch]
#     # labels = [x["cuis"] for x in batch]
#     labels = [x["db_ids"] for x in batch]
#     metadata = batch

#     return mentions, labels, metadata


class MetricLearningDataset_pairwise(Dataset):
    """
    Candidate Dataset for:
        query_tokens, candidate_tokens, label
    """
    def __init__(
            self, 
            training_pairs: List[str], 
            tokenizer): #d_ratio, s_score_matrix, s_candidate_idxs):
        self.query_ids = []
        self.query_names = []
        for i, line in tqdm(enumerate(training_pairs), desc="Loading training pairs"):
            line = line.rstrip("\n")
            if (len(line.split("||")) == 3):
                query_id, name1, name2 = line.split("||")
                self.query_ids.append(query_id)
                self.query_names.append((name1, name2))
        print("Lengths of the dataset: ", len(self.query_ids), len(self.query_names))
        self.tokenizer = tokenizer
        self.query_id_2_index_id = {k: v for v, k in enumerate(list(set(self.query_ids)))}
    
    def __getitem__(self, query_idx):

        query_name1 = self.query_names[query_idx][0]
        query_name2 = self.query_names[query_idx][1]
        query_id = self.query_ids[query_idx]
        query_id = int(self.query_id_2_index_id[query_id])

        return query_name1, query_name2, query_id


    def __len__(self):
        return len(self.query_names)



class MetricLearningDataset(Dataset):
    """
    Candidate Dataset for:
        query_tokens, candidate_tokens, label
    """
    def __init__(self, path, tokenizer): #d_ratio, s_score_matrix, s_candidate_idxs):
        logger.info("Initializing metric learning data set! ...")
        with open(path, 'r') as f:
            lines = f.readlines()
        self.query_ids = []
        self.query_names = []
        cuis = []
        for line in lines:
            cui, _ = line.split("||")
            cuis.append(cui)

        self.cui2id = {k: v for v, k in enumerate(cuis)}
        for line in lines:
            line = line.rstrip("\n")
            cui, name = line.split("||")
            query_id = self.cui2id[cui]
            #if query_id.startswith("C"):
            #    query_id = query_id[1:]
            #query_id = int(query_id)
            self.query_ids.append(query_id)
            self.query_names.append(name)
        self.tokenizer = tokenizer
    
    def __getitem__(self, query_idx):

        query_name = self.query_names[query_idx]
        query_id = self.query_ids[query_idx]
        query_token = self.tokenizer.transform([query_name])[0]

        return torch.tensor(query_token), torch.tensor(query_id)

    def __len__(self):
        return len(self.query_names)


if __name__ == "__main__":
    dataset = SapBertBigBioDataset(
        path_to_abbreviation_dict="/home/pbathala3/entity_linking/biomedical-entity-linking/bioel/bioel/utils/solve_abbreviation/abbreviations.json"
    )
    #dataloader = DataLoader(dataset, collate_fn=sapbert_collate_fn, batch_size=1)
    # for i, batch in enumerate(tqdm(dataloader)):
    #     if i < 1:
    #         print(batch)
    
    ontology_config = {
        "filepath": "/mitchell/entity-linking/2022AA/META/",
        "name": "UMLS",
        "abbrev": None,
    }
    ontology = BiomedicalOntology.load_umls(**ontology_config)
    pos_pairs = generate_positive_pairs(ontology.entities)
    for i, pair in enumerate(tqdm(pos_pairs)):
        if i < 1:
            print(pair)


import argparse
import os
import glob
from typing import List, Tuple, Dict, Iterator, Set
import torch
import numpy as np
import ujson
from bioel.models.krissbert.model.model import Krissbert
from bioel.models.krissbert.data.utils import BigBioDataset
from transformers import set_seed

import logging

# Setup logger
logger = logging.getLogger()
logger.setLevel(logging.INFO)
log_formatter = logging.Formatter(
    "[%(thread)s] %(asctime)s [%(levelname)s] %(name)s: %(message)s"
)
console = logging.StreamHandler()
console.setFormatter(log_formatter)
logger.addHandler(console)


def load_umls_data(files_patterns: List[str], candidate_ids: Dict = None) -> Dict:
    input_paths = []
    for pattern in files_patterns:
        pattern_files = glob.glob(pattern)
        input_paths.extend(pattern_files)
    umls_data = {}
    for file in sorted(input_paths):
        logger.info("Reading encoded UMLS data from file %s", file)
        with open(file, "rb") as reader:
            for meta, vec in pickle.load(reader):
                assert len(meta["cuis"]) == 1, breakpoint()
                cui = meta["cuis"][0]
                if candidate_ids and cui not in candidate_ids:
                    continue
                umls_data[cui] = (meta, vec)
    logger.info(f"Loaded UMLS data = {len(umls_data)}.")
    return umls_data


def hit(pred: List[str], gold: List[str]) -> bool:
    return all(p in gold for p in pred)


def dedup_ids(ids: List[Dict]) -> List[Dict]:
    deduped_ids = []
    seen_cuis = set()
    for d in ids:
        if all(cui in seen_cuis for cui in d["cuis"]):
            continue
        seen_cuis.update(d["cuis"])
        deduped_ids.append(d)
    return deduped_ids


def evaluate(
    ds: torch.utils.data.Dataset,
    result_ent_ids: List[Tuple[List[object], List[float]]],
    lookup_table: str,
    output_path: str,
    top_ks: List[int] = (1, 5, 50, 100),
    path_to_abbrev: str = None,
) -> List[Dict]:
    """
    Evaluate model
    """
    # Load lookup table
    # Each entity name is mapped to
    lut = {}
    with open(lookup_table, encoding="utf-8") as f:
        for ln in f:
            cuis, name = ln.strip().split("||")
            cuis = cuis.split("|")
            lut[name] = cuis

    n = len(ds)
    top_k_hits = {top_k: 0 for top_k in top_ks}

    all_output = []

    for i in range(len(result_ent_ids)):
        d = ds[i]
        ids, _ = result_ent_ids[i]
        # logger.info(ids)
        ids = dedup_ids(ids)
        ids = ids[: max(top_ks)]
        candidates = [
            {"cuis": eid["cuis"], "hit": int(hit(pred=eid["cuis"], gold=d.cuis))}
            for eid in ids
        ]
        lut_cuis = lut.get(d.mention, [])
        if len(lut_cuis) == 1:
            # If the mention only has one ID in the look up table,
            # we use the ID as the top prediction.
            candidates.insert(
                0, {"cuis": lut_cuis, "hit": int(hit(pred=lut_cuis, gold=d.cuis))}
            )

        output_candidates = [x["cuis"] for x in candidates]
        output = d.to_dict()
        # print(output)
        output["candidates"] = output_candidates
        if path_to_abbrev is not None:
            output["mention_id"] = str(output["mention_id"]) + ".abbr_resolved"
        # output['hits'] = [x['hit'] for x in candidates]
        all_output.append(output)

        for top_k in top_k_hits:
            if any(c["hit"] for c in candidates[:top_k]):
                top_k_hits[top_k] += 1

    with open(output_path, "w") as f:
        f.write(ujson.dumps(all_output, indent=2))

    top_k_acc = {top_k: v / n for top_k, v in top_k_hits.items()}
    logger.info("Top-k accuracy %s", top_k_acc)


def evaluate_model(params, model):
    top_ks = [1, 5, 50, 100]
    encoded_umls_files = []
    encoded_files = []
    entity_list_names = None
    entity_list_ids = None

    set_seed(params["seed"])

    if "device" in params:
        os.environ["CUDA_VISIBLE_DEVICES"] = str(params["device"])
        print(f"Set CUDA_VISIBLE_DEVICES to {params['device']}")

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print(f"Using device: {device}")
    model.to(device)

    # Load the dataset
    ds = BigBioDataset(
        params["dataset_name"],
        splits=["test"],
        abbreviations_path=params["path_to_abbrev"],
    )

    # candidate ids
    candidate_ids = None
    if entity_list_ids:
        with open(entity_list_ids, encoding="utf-8") as f:
            candidate_ids = set(f.read().split("\n"))

    # Start indexing
    input_paths = []
    for pattern in encoded_files:
        pattern_files = glob.glob(pattern)
        input_paths.extend(pattern_files)
    input_paths = sorted(set(input_paths))

    if len(input_paths) == 0:
        input_paths.append(params["input_path"])

    mentions_tensor = model.generate_mention_vectors(ds)

    # Load UMLS knowledge
    umls_data = None
    if encoded_umls_files:
        umls_data = load_umls_data(encoded_umls_files, candidate_ids)

    index_path = params["index_path"]
    if index_path and model.index.index_exists(index_path):
        logger.info("Index path: %s", index_path)
        model.index.deserialize(index_path)
    else:
        logger.info("Indexing encoded data from files: %s", input_paths)
        model.index_encoded_data(
            vector_files=input_paths,
            buffer_size=model.index.buffer_size,
            candidate_ids=candidate_ids,
            umls_data=umls_data,
        )
        if index_path:
            os.makedirs(os.path.dirname(index_path), exist_ok=True)
            model.index.serialize(index_path)

    # Encode test data.
    mentions_tensor = torch.cat([mentions_tensor, mentions_tensor], dim=1)

    # To get k different entities, we retrieve 32 * k mentions and then dedup.
    top_ids_and_scores = model.get_top_hits(
        mentions_tensor.numpy(), params["num_retrievals"] * 32
    )

    if entity_list_names:
        entity_list_names = entity_list_names
    else:
        entity_list_names = params["entity_list_names"]

    output_path = params["output_path"]
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    evaluate(
        ds=ds,
        result_ent_ids=top_ids_and_scores,
        lookup_table=entity_list_names,
        output_path=output_path,
        top_ks=top_ks,
        path_to_abbrev=params["path_to_abbrev"],
    )



from bioel.models.biogenel.fairseq_beam.sequence_generator import (
    FairseqIncrementalDecoder,
    SequenceGenerator,
    EnsembleModel,
)
from bioel.models.biogenel.fairseq_beam.search import (
    Search,
    BeamSearch,
    PrefixConstrainedBeamSearch,
    PrefixConstrainedBeamSearchWithSampling,
)
from bioel.models.biogenel.fairseq_beam.sequence_scorer import sequence_score



import os
import sys
# import Path
import numpy as np
import tqdm
import pickle
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
# from transformers import BartForConditionalGeneration
from bioel.models.biogenel.models.bart_model import BartForConditionalGeneration

from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from transformers.file_utils import ModelOutput
from transformers.generation.logits_process import (
    EncoderNoRepeatNGramLogitsProcessor,
    ForcedBOSTokenLogitsProcessor,
    ForcedEOSTokenLogitsProcessor,
    HammingDiversityLogitsProcessor,
    InfNanRemoveLogitsProcessor,
    PrefixConstrainedLogitsProcessor,
    LogitsProcessorList,
    MinLengthLogitsProcessor,
    NoBadWordsLogitsProcessor,
    NoRepeatNGramLogitsProcessor,
    RepetitionPenaltyLogitsProcessor,
    TemperatureLogitsWarper,
    TopKLogitsWarper,
    TopPLogitsWarper,
)
from transformers.generation.stopping_criteria import (
    MaxLengthCriteria,
    MaxNewTokensCriteria,
    MaxTimeCriteria,
    StoppingCriteriaList,
    validate_stopping_criteria,
)
from transformers.utils import logging
from transformers.generation.utils import *
from bioel.models.biogenel.models.beam_scorer import BeamScorer, BeamSearchScorer


class BartEntityPromptModel(BartForConditionalGeneration):

    def __init__(self, config):
        super().__init__(config)

    @classmethod
    def from_pretrained(
        cls, 
        pretrained_model_name_or_path: str,
        config,
        vocabs: list = [[],[]],
        finetune: bool = False,
        n_tokens: tuple = (10, 10),
        load_prompt: bool = False,
        soft_prompt_path: str = None, 
        initialize_from_vocab: bool = False,
        random_range: float = 0.5,
        no_finetune_decoder: bool=False,
        **kwargs,
    ):

        model = super().from_pretrained(pretrained_model_name_or_path, config = config)

        if not finetune:
            print('no finetune!')
            for param in model.parameters():
                param.requires_grad = False

        if no_finetune_decoder:
            print('not finetune decoder!')
            for param in model.model.decoder.parameters():
                param.requires_grad = False
            for param in model.model.encoder.embed_tokens.parameters():
                param.requires_grad = False
            for param in model.model.shared.parameters():
                param.requires_grad = False
            for param in model.lm_head.parameters():
                param.requires_grad = False
            
        #initialize soft prompt
        model.set_prompt_number(n_tokens)

        if n_tokens[0] > 0 and n_tokens[1] > 0:
            print("Initializing soft prompt...")
            model.initialize_soft_prompt(
                n_tokens=n_tokens,
                initialize_from_vocab=initialize_from_vocab,
                vocabs=vocabs,
                random_range=random_range,
            )
            #load soft prompts
            if load_prompt:
                print('Loading soft prompts.....')
                model.set_soft_prompt_embeds(soft_prompt_path)
        else:
            print('Setting no soft prompts!')

        return model

    def set_prompt_number(
        self,
        n_tokens: tuple,
        ):
        self.n_tokens_enc = n_tokens[0]
        self.n_tokens_dec = n_tokens[1]
        self.soft_prompt_enc = None
        self.soft_prompt_dec = None

    def set_soft_prompt_embeds(
        self,
        soft_prompt_path: str,
    ):
        params = torch.load(os.path.join(soft_prompt_path, 'pytorch_model.bin'))
        if 'soft_prompt_enc.weight' in params:
            self.soft_prompt_enc.weight.data = params['soft_prompt_enc.weight']
            self.soft_prompt_dec.weight.data = params['soft_prompt_dec.weight']
            print('Load from pytorch_model.bin')
        elif os.path.exists(os.path.join(soft_prompt_path, 'soft_prompt_enc.pth')):
            self.soft_prompt_enc = torch.load(os.path.join(soft_prompt_path, 'soft_prompt_enc.pth'), map_location=torch.device(self.device))
            self.soft_prompt_dec = torch.load(os.path.join(soft_prompt_path, 'soft_prompt_dec.pth'), map_location=torch.device(self.device))
        else:
            print('Find no soft prompt parameter in the path')
        print(f"Loaded encoder soft prompt! (n_tokens: {self.n_tokens_enc})")
        print(f"Loaded decoder soft prompt! (n_tokens: {self.n_tokens_dec})")

    def save_soft_prompt(
        self, 
        path: str
        ):
        # Path(path).mkdir(parents=True, exist_ok=True)
        torch.save(self.soft_prompt_enc, os.path.join(path, 'soft_prompt_enc.pth'))
        torch.save(self.soft_prompt_dec, os.path.join(path, 'soft_prompt_dec.pth'))

    def initialize_soft_prompt(
        self,
        n_tokens: tuple,
        initialize_from_vocab: bool,
        vocabs: list,
        random_range: float = 0.5,
    ):
        if initialize_from_vocab:
            init_prompt_value_enc = self.model.encoder.embed_tokens(vocabs[0]).data
            init_prompt_value_dec = self.model.decoder.embed_tokens(vocabs[1]).data
        else:
            init_prompt_value_enc = torch.FloatTensor(self.n_tokens_enc, self.config.d_model).uniform_(
                -random_range, random_range
            )
            init_prompt_value_dec = torch.FloatTensor(self.n_tokens_dec, self.config.d_model).uniform_(
                -random_range, random_range
            )
        self.soft_prompt_enc = nn.Embedding(self.n_tokens_enc, self.config.d_model)
        self.soft_prompt_dec = nn.Embedding(self.n_tokens_dec, self.config.d_model)
        print(f"Set encoder soft prompt! (n_tokens: {self.n_tokens_enc})")
        print(f"Set decoder soft prompt! (n_tokens: {self.n_tokens_dec})")
    
    def id2emb_cat_prompt(self, input_ids, coder_part = 'encoder', insert_idx = None):
        
        if coder_part == 'encoder':
            coder = self.model.encoder
            n_tokens = self.n_tokens_enc
            prompt_embeds = self.soft_prompt_enc
        else:
            coder = self.model.decoder
            n_tokens = self.n_tokens_dec
            prompt_embeds = self.soft_prompt_dec

        inputs_embeds = coder.embed_tokens(input_ids)
        if len(list(inputs_embeds.shape)) == 2:
            inputs_embeds = inputs_embeds.unsqueeze(0)

        if n_tokens > 0:
            prompt_embeds = prompt_embeds.weight.repeat(input_ids.size(0), 1, 1)
            if insert_idx is not None:
                inputs = []
                for i in range(inputs_embeds.shape[0]):
                    inputs.append(torch.cat([inputs_embeds[i,0:insert_idx[i][0],:], 
                                                prompt_embeds[i,:n_tokens//2,:], 
                                                inputs_embeds[i,insert_idx[i][0]:insert_idx[i][1],:],
                                                prompt_embeds[i,n_tokens//2:,:], 
                                                inputs_embeds[i,insert_idx[i][1]:,:]], dim=0) )
                inputs_embeds = torch.stack(inputs)

            else:
                inputs_embeds = torch.cat([prompt_embeds, inputs_embeds], dim=1)
                
        return inputs_embeds.to(self.device)
    
    def extend_attention_mask(
        self, 
        attention_mask,
        n_tokens,
        ):

        if len(list(attention_mask.shape)) == 1:
            attention_mask = attention_mask.unsqueeze(0)
        if n_tokens > 0:
            attention_mask = torch.cat([torch.full((attention_mask.shape[0], n_tokens), 1).to(self.device), attention_mask],dim=1)

        return attention_mask.to(self.device)

    def forward(
        self,
        input_ids=None,
        decoder_input_ids=None,
        labels=None,
        attention_mask=None,
        decoder_attention_mask=None,
        inputs_embeds=None,
        decoder_inputs_embeds=None,
        encoder_outputs=None,
        use_cache=None,
        return_dict=None,
        mention_indexes=None,
        **kwargs,
        ):

        if encoder_outputs is None:
            if input_ids is not None:
                inputs_embeds_enc = self.id2emb_cat_prompt(input_ids, coder_part = 'encoder', insert_idx = mention_indexes)
            else:
                inputs_embeds_enc = inputs_embeds.to(self.device)
            
            if decoder_input_ids is not None:
                inputs_embeds_dec = self.id2emb_cat_prompt(decoder_input_ids, coder_part = 'decoder')
            else:
                inputs_embeds_dec = decoder_inputs_embeds.to(self.device)
            
            if attention_mask is not None:
                attention_mask_enc = self.extend_attention_mask(attention_mask, self.n_tokens_enc)
            else:
                attention_mask_enc = attention_mask
            
            if decoder_attention_mask is not None:
                attention_mask_dec = self.extend_attention_mask(decoder_attention_mask, self.n_tokens_dec)
            else:
                attention_mask_dec = decoder_attention_mask
            
            return super().forward(

                attention_mask=attention_mask_enc,
                decoder_attention_mask=attention_mask_dec,

                inputs_embeds=inputs_embeds_enc,
                decoder_inputs_embeds=inputs_embeds_dec,

                labels=labels,

                use_cache=use_cache,
                return_dict=return_dict,

                **kwargs,
            )
                
        else:
            
            if decoder_input_ids is not None:
                inputs_embeds_dec = self.id2emb_cat_prompt(decoder_input_ids, coder_part = 'decoder')
            else:
                inputs_embeds_dec = decoder_inputs_embeds.to(self.device)

            if decoder_attention_mask is not None:
                attention_mask_dec = self.extend_attention_mask(decoder_attention_mask, self.n_tokens_dec)
            else:
                attention_mask_dec = decoder_attention_mask

            return super().forward(
                input_ids=None,
                decoder_input_ids=None,

                attention_mask=None,
                decoder_attention_mask=attention_mask_dec,

                inputs_embeds=None,
                decoder_inputs_embeds=inputs_embeds_dec,
                encoder_outputs=encoder_outputs,

                use_cache=use_cache,
                return_dict=return_dict,

                **kwargs,
            )



import logging

def setup_logger():
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    if not logger.handlers:  # Check if the logger already has handlers
        logger.setLevel(logging.DEBUG)

        formatter = logging.Formatter(
            "[%(asctime)s] [%(filename)s] [%(levelname)s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)

        logger.addHandler(stream_handler)
        logger.propagate = False

    return logger 


from typing import List, Dict
import time
import logging
import json
from dataclasses import dataclass
import ujson

import torch
from torch import Tensor as T
from transformers import PreTrainedTokenizer
from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    dataset_to_df,
    resolve_abbreviation,
    dataset_to_documents,
    get_left_context,
    get_right_context,
    add_deabbreviations,
)
from bioel.utils.dataset_consts import (
    DATASET_NAMES,
    CUIS_TO_EXCLUDE,
    CUIS_TO_REMAP,
)
from tqdm.auto import tqdm

tqdm.pandas()


logger = logging.getLogger()


@dataclass
class Mention:
    cui: str
    start: int
    end: int
    text: str
    types: str
    deabbreviated_text: str


@dataclass
class ContextualMention:
    mention: str  # text
    cuis: List[str]
    ctx_l: str
    ctx_r: str
    deabbreviated_text: str

    def to_tensor(self, tokenizer: PreTrainedTokenizer, max_length: int) -> T:
        ctx_l_ids = tokenizer.encode(
            text=self.ctx_l,
            add_special_tokens=False,
            max_length=max_length,
            truncation=True,
        )
        ctx_r_ids = tokenizer.encode(
            text=self.ctx_r,
            add_special_tokens=False,
            max_length=max_length,
            truncation=True,
        )
        mention_ids = tokenizer.encode(
            text=self.mention,
            add_special_tokens=False,
            max_length=max_length,
            truncation=True,
        )

        # Concatenate context and mention to the max length.
        token_ids = (
            tokenizer.convert_tokens_to_ids(["<ENT>"])
            + mention_ids
            + tokenizer.convert_tokens_to_ids(["</ENT>"])
        )
        max_ctx_len = max_length - len(token_ids) - 2  # Exclude [CLS] and [SEP]
        max_ctx_l_len = max_ctx_len // 2
        max_ctx_r_len = max_ctx_len - max_ctx_l_len
        if len(ctx_l_ids) < max_ctx_l_len and len(ctx_r_ids) < max_ctx_r_len:
            token_ids = ctx_l_ids + token_ids + ctx_r_ids
        elif len(ctx_l_ids) >= max_ctx_l_len and len(ctx_r_ids) >= max_ctx_r_len:
            token_ids = (
                ctx_l_ids[-max_ctx_l_len:] + token_ids + ctx_r_ids[:max_ctx_r_len]
            )
        elif len(ctx_l_ids) >= max_ctx_l_len:
            ctx_l_len = max_ctx_len - len(ctx_r_ids)
            token_ids = ctx_l_ids[-ctx_l_len:] + token_ids + ctx_r_ids
        else:
            ctx_r_len = max_ctx_len - len(ctx_l_ids)
            token_ids = ctx_l_ids + token_ids + ctx_r_ids[:ctx_r_len]

        token_ids = [tokenizer.cls_token_id] + token_ids

        # The above snippet doesn't guarantee the max length limit.
        token_ids = token_ids[: max_length - 1] + [tokenizer.sep_token_id]

        if len(token_ids) < max_length:
            token_ids = token_ids + [tokenizer.pad_token_id] * (
                max_length - len(token_ids)
            )

        return torch.tensor(token_ids)


@dataclass
class DetailedContextualMention(ContextualMention):
    document_id: str
    types: str
    offsets: List[List[int]]
    # start: int
    # end: int
    mention_id: str
    split: str

    def to_dict(self):
        return {
            "document_id": self.document_id,
            "offsets": self.offsets,
            "text": self.mention,
            "type": self.types,
            # "start": self.start,
            # "end": self.end,
            "db_ids": self.cuis,
            "split": self.split,
            "deabbreviated_mention": self.deabbreviated_text,
            "mention_id": self.mention_id,
        }


class BigBioDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        dataset_name: str,
        splits: List[str],
        abbreviations_path=None,
    ):
        self.dataset_name = dataset_name
        self.data = load_bigbio_dataset(dataset_name)
        self.splits = splits
        self.abbreviations_path = abbreviations_path
        self.name_to_cuis = {}

        exclude = CUIS_TO_EXCLUDE[dataset_name]
        remap = CUIS_TO_REMAP[dataset_name]

        if self.abbreviations_path:
            self.data = add_deabbreviations(
                dataset=self.data, path_to_abbrev=self.abbreviations_path
            )
            print("Resolved abbreviations")
        else:
            print("No abbreviations dictionary found.")

        df = dataset_to_df(
            self.data, cuis_to_exclude=exclude, entity_remapping_dict=remap
        )
        df["start"] = df["offsets"].map(lambda x: x[0][0])
        df["end"] = df["offsets"].map(lambda x: x[-1][-1])
        df = df[df.split.isin(splits)]
        print(dataset_name, splits, df.split.unique())

        self.df = df

        self.documents = dataset_to_documents(self.data)

        self._post_init()

    def _df_to_contextual_mentions(self, max_length: int = 64):
        self.df["ctx_l"] = self.df[["document_id", "start"]].progress_apply(
            lambda x: get_left_context(
                self.documents[x.iloc[0]], x.iloc[1], strip=True
            ),
            axis=1,
        )
        self.df["ctx_r"] = self.df[["document_id", "end"]].progress_apply(
            lambda x: get_right_context(
                self.documents[x.iloc[0]], x.iloc[1], strip=True
            ),
            axis=1,
        )

        self.df = self.df.rename(
            {"db_ids": "cuis", "type": "types", "text": "mention"}, axis=1
        ).drop(["start", "end"], axis=1)

        return [
            DetailedContextualMention(**x) for x in self.df.to_dict(orient="records")
        ]

    def to_contextual_mentions(self):
        """
        Turn dataset into set of contextual mentions
        """
        self.mentions = self._df_to_contextual_mentions()

    def _post_init(self):
        self.to_contextual_mentions()

        for m in self.mentions:
            if m.mention not in self.name_to_cuis:
                self.name_to_cuis[m.mention] = set()
            self.name_to_cuis[m.mention].update(m.cuis)

    def __len__(self):
        return len(self.mentions)

    def __getitem__(self, idx: int) -> DetailedContextualMention:
        return self.mentions[idx]


def split_dataset(docs: List, split_path_prefix: str) -> Dict[str, List]:
    split_kv = {"train": "trng", "dev": "dev", "test": "test"}
    id_to_split = {}
    dataset = {}
    for k, v in split_kv.items():
        dataset[k] = []
        path = split_path_prefix + v + ".txt"
        for i in open(path, encoding="utf-8").read().strip().split("\n"):
            assert i not in id_to_split, breakpoint()
            id_to_split[i] = k
    for doc in docs:
        split = id_to_split[doc.id]
        dataset[split].append(doc)
    return dataset


class PreprocessedDataset(torch.utils.data.Dataset):
    def __init__(self, dataset_path: str) -> None:
        super().__init__()
        self.file = dataset_path
        self.data = []
        self.load_data()

    def load_data(self) -> None:
        with open(self.file, encoding="utf-8") as f:
            logger.info("Reading file %s" % self.file)
            for ln in f:
                if ln.strip():
                    self.data.append(json.loads(ln))
        logger.info("Loaded data size: {}".format(len(self.data)))

    def __getitem__(self, index: int) -> ContextualMention:
        d = self.data[index]
        return ContextualMention(
            ctx_l=d["context_left"],
            ctx_r=d["context_right"],
            mention=d["mention"],
            cuis=d["cuis"],
        )

    def __len__(self) -> int:
        return len(self.data)


def generate_vectors(
    encoder: torch.nn.Module,
    tokenizer: PreTrainedTokenizer,
    dataset: torch.utils.data.Dataset,
    batch_size: int,
    max_length: int,
    is_prototype: bool = False,
):
    n = len(dataset)
    total = 0
    results = []
    start_time = time.time()
    logger.info("Start encoding...")
    for i, batch_start in enumerate(range(0, n, batch_size)):
        batch = [
            dataset[i] for i in range(batch_start, min(n, batch_start + batch_size))
        ]
        batch_token_tensors = [m.to_tensor(tokenizer, max_length) for m in batch]

        ids_batch = torch.stack(batch_token_tensors, dim=0).cuda()
        seg_batch = torch.zeros_like(ids_batch)
        attn_mask = ids_batch != tokenizer.pad_token_id

        with torch.inference_mode():
            out = encoder(
                input_ids=ids_batch, token_type_ids=seg_batch, attention_mask=attn_mask
            )
            out = out[0][:, 0, :]
        out = out.cpu()

        num_mentions = out.size(0)
        total += num_mentions

        if is_prototype:
            meta_batch = [{"cuis": m.cuis} for m in batch]
            assert len(meta_batch) == num_mentions
            results.extend(
                [(meta_batch[i], out[i].view(-1).numpy()) for i in range(num_mentions)]
            )
        else:
            results.extend(out.cpu().split(1, dim=0))

        if (i + 1) % 10 == 0:
            eta = (n - total) * (time.time() - start_time) / 60 / total
            logger.info(f"Batch={i + 1}, Encoded mentions={total}, ETA={eta:.1f}m")

    assert len(results) == n
    logger.info(f"Total encoded mentions={n}")
    if not is_prototype:
        results = torch.cat(results, dim=0)

    return results


if __name__ == "__main__":
    # Tests
    data = BigBioDataset("medmentions_full", splits=["train"])
    print(len(data))

    data = BigBioDataset("medmentions_full", splits=["train", "validation"])
    print(len(data))



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import argparse
import json
import ujson
import logging
import os
import torch
import pickle
from tqdm import tqdm

from torch.utils.data import DataLoader, SequentialSampler

from transformers import AutoTokenizer

from bioel.models.arboel.model.biencoder import BiEncoderRanker
import bioel.models.arboel.data.data_process as data_process
from bioel.models.arboel.crossencoder.original.train_cross import (
    read_dataset,
)
from bioel.models.arboel.model.common.params import BlinkParser

from IPython import embed

# Configure the logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_topk_predictions(
    input_crossencoder,
    reranker,
    train_dataloader,
    candidate_pool,
    cand_encode_list,
    silent,
    logger,
    top_k=10,
    save_predictions=False,
):
    """
    Retrieves the top-k predictions for a set of inputs using a reranking model.

    ------
    Params:
    - reranker : model
    biencoder
    - train_dataloader : DataLoader
    Provides batches of input data for evaluation.
    - candidate_pool : list of list (dim : )
    Contains the IDs of the candidate description.
    - cand_encode_list : list
    A list of precomputed candidate encodings that the reranker uses to score the inputs.
    - silent (bool): If True, the function operates silently without outputting progress; if False, progress
      is displayed using tqdm.
    - logger : logger object
    - top_k : int
    The number of top predictions to retrieve for each input, default is 10.
    - save_predictions : bool
    If True, saves the context, candidate indices, and label information for each prediction.

    Returns:
    - dict: A dictionary containing tensors for the contexts, candidate vectors, and labels of the top-k predictions.
      This includes:
        - "context_vecs": Tensor of input contexts.
        - "candidate_vecs": Tensor of candidate data corresponding to the top predictions.
        - "labels": Tensor of label indices indicating the position of the correct candidate within the top-k.
    """
    reranker.model.eval()
    logger.info("Getting top %d predictions." % top_k)
    if silent:
        iter_ = train_dataloader
    else:
        iter_ = tqdm(train_dataloader)

    count = 0
    n_samples = 0

    nn_context = []
    nn_candidates = []
    nn_labels = []
    nn_labels_bis = []

    indicies = input_crossencoder
    print("len(indicies) :", len(indicies))
    assert len(indicies) != 0, "list of top k candidates idx is empty"
    assert (
        len(indicies[0]) >= top_k
    ), f"{len(indicies[0])} <= {top_k} : Dimension mismatch between number of candidates (top_k) and the length of the list of top candidates idx"

    for step, batch in enumerate(iter_):
        length = len(batch)
        batch = tuple(t.cuda() for t in batch)  # Multiple tensors
        if len(batch) == 4:
            context_input, _, srcs, label_ids = batch

        elif len(batch) == 3:
            context_input, _, label_ids = batch

        for i in range(context_input.size(0)):

            inds = indicies[n_samples]
            n_samples += 1

            pointer = -1
            for j in range(top_k):
                if inds[j] == label_ids[i].item():
                    pointer = j  #  position of the gold cui in the candidate list
                    break

            if pointer == -1:
                count += 1
                continue

            if not save_predictions:
                continue

            # add examples in new_data
            cur_candidates = candidate_pool[inds]
            nn_context.append(context_input[i].cpu().tolist())
            nn_candidates.append(cur_candidates[:top_k].cpu().tolist())
            nn_labels.append(pointer)
            nn_labels_bis.append(label_ids[i].item())

    print("total_correct_samples:", n_samples - count)
    print("total_samples:", n_samples)
    nn_context = torch.LongTensor(nn_context)
    nn_candidates = torch.LongTensor(nn_candidates)
    nn_labels = torch.LongTensor(nn_labels)
    nn_labels_bis = torch.LongTensor(nn_labels_bis)
    nn_data = {
        "context_vecs": nn_context,
        "candidate_vecs": nn_candidates,
        "labels": nn_labels,
        "labels_bis": nn_labels_bis,
    }

    return nn_data


def load_entity_dict(logger, params):
    """
    Loads and processes a dictionary of entities from a specified file path.
    ------
    Params:
    - logger : logger object
    - params : dict
    A dictionary of parameters that control the function's behavior
    ------
    Returns:
    - list of tuples: Each tuple contains two elements, the 'title' and the 'text' (description) of an entity.
    """
    path = params.get("entity_dict_path", None)
    assert path is not None, "Error! entity_dict_path is empty."

    logger.info("Loading entity description from path: " + path)
    entity_list = []  # Initialize the list to store tuples of title and text
    with open(path, "rb") as f:  # Open file in binary read mode
        full_list = pickle.load(f)

        for item in full_list:
            title = item["title"]
            text = item.get("description", "").strip()
            entity_list.append((title, text))

            if params.get("debug", False) and len(entity_list) > 200:
                break

    return entity_list


def get_candidate_pool_tensor(
    entity_desc_list,
    tokenizer,
    max_seq_length,
    logger,
):
    """
    Processes a list of entity descriptions by converting them into token IDs.
    The resulting IDs are then collected into a tensor.
    ------
    Params:
    - entity_desc_list : list of str or tuple
    A list where each element is either a string representing an entity description or a tuple containing a title and an entity description.
    - tokenizer : PreTrainedTokenizer
    Used to convert text into a sequence of token IDs.
    - max_seq_length : int
    The maximum length of the token sequences.
    - logger : logger object

    Returns:
    - torch.LongTensor: Contains the token IDs of the entity descriptions.
    """
    # TODO: add multiple thread process
    logger.info("Convert candidate text to id")
    cand_pool = []
    for entity_desc in tqdm(entity_desc_list):
        if type(entity_desc) is tuple:
            title, entity_text = entity_desc
        else:
            title = None
            entity_text = entity_desc

        rep = data_process.get_candidate_representation(
            entity_text,
            tokenizer,
            max_seq_length,
            title,
        )
        cand_pool.append(rep["ids"])

    cand_pool = torch.LongTensor(cand_pool)
    return cand_pool


# def encode_candidate(
#     reranker,
#     candidate_pool,
#     encode_batch_size,
#     silent,
#     logger,
# ):
#     """
#     Encodes a pool of candidate entries using a reranker model.
#     ------
#     Params:
#     - reranker : model
#     biencoder
#     - candidate_pool : Dataset
#     A dataset containing candidate entries that need to be encoded.
#     - encode_batch_size : int
#     Size of the batches in which the candidate_pool will be processed.
#     - silent : bool
#     If True, the function operates without printing progress; if False, progress is displayed using tqdm.
#     - logger : logger object

#     Returns:
#     - torch.Tensor: Contains the encoded representations of all candidates in the candidate pool.
#     """
#     reranker.model.eval()
#     sampler = SequentialSampler(candidate_pool)
#     data_loader = DataLoader(
#         candidate_pool, sampler=sampler, batch_size=encode_batch_size
#     )
#     if silent:
#         iter_ = data_loader
#     else:
#         iter_ = tqdm(data_loader)

#     cand_encode_list = None
#     for step, batch in enumerate(iter_):
#         cand_encode = reranker.encode_candidate(batch.cuda())
#         if cand_encode_list is None:
#             cand_encode_list = cand_encode
#         else:
#             cand_encode_list = torch.cat((cand_encode_list, cand_encode))

#     return cand_encode_list


def load_or_generate_candidate_pool(
    tokenizer,
    params,
    logger,
    cand_pool_path,
):
    candidate_pool = None
    if cand_pool_path is not None:
        # try to load candidate pool from file
        try:
            logger.info("Loading pre-generated candidate pool from: ")
            logger.info(cand_pool_path)
            candidate_pool = torch.load(cand_pool_path)
        except:
            logger.info("Loading failed. Generating candidate pool")

    if candidate_pool is None:
        # compute candidate pool from entity list
        entity_desc_list = load_entity_dict(logger, params)
        candidate_pool = get_candidate_pool_tensor(
            entity_desc_list,
            tokenizer,
            params["max_cand_length"],
            logger,
        )

        if cand_pool_path is not None:
            logger.info("Saving candidate pool.")
            torch.save(candidate_pool, cand_pool_path)

    return candidate_pool


def eval_biencoder(params, reranker, tokenizer, input_crossencoder):
    output_path = params["output_path"]
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    model = reranker.model

    cand_encode_path = params.get("cand_encode_path", None)

    # candidate encoding is not pre-computed.
    # load/generate candidate pool to compute candidate encoding.
    cand_pool_path = params.get("cand_pool_path", None)
    candidate_pool = load_or_generate_candidate_pool(
        tokenizer,
        params,
        logger,
        cand_pool_path,
    )

    candidate_encoding = None
    if cand_encode_path is not None:
        # try to load candidate encoding from path
        # if success, avoid computing candidate encoding
        try:
            logger.info("Loading pre-generated candidate encode path.")
            candidate_encoding = torch.load(cand_encode_path)
        except:
            logger.info("Loading failed. Generating candidate encoding.")

    if candidate_encoding is None:
        candidate_encoding = data_process.embed_and_index(
            model=reranker,
            token_id_vecs=candidate_pool,
            encoder_type="candidate",
            batch_size=768,
            only_embed=True,
        )

        if cand_encode_path is not None:
            # Save candidate encoding to avoid re-compute
            logger.info("Saving candidate encoding to file " + cand_encode_path)
            torch.save(candidate_encoding, cand_encode_path)

    test_samples = read_dataset(params["mode"], params["data_path"])
    logger.info("Read %d test samples." % len(test_samples))

    entity_dictionary_pkl_path = os.path.join(
        params["data_path"], "entity_dictionary.pickle"
    )
    entity_dictionary_loaded = False
    if os.path.isfile(entity_dictionary_pkl_path):
        print("Loading stored processed entity dictionary...")
        with open(entity_dictionary_pkl_path, "rb") as read_handle:
            entity_dictionary = pickle.load(read_handle)
        entity_dictionary_loaded = True

    test_data, test_tensor_data = data_process.process_mention_with_candidate(
        samples=test_samples,
        entity_dictionary=entity_dictionary,
        tokenizer=tokenizer,
        max_context_length=params["max_context_length"],
        max_cand_length=params["max_cand_length"],
        context_key=params["context_key"],
        silent=params["silent"],
        dictionary_processed=entity_dictionary_loaded,
        logger=logger,
        debug=params["debug"],
    )

    test_sampler = SequentialSampler(test_tensor_data)
    test_dataloader = DataLoader(
        test_tensor_data, sampler=test_sampler, batch_size=params["encode_batch_size"]
    )

    save_results = params.get("save_topk_result")
    new_data = get_topk_predictions(
        input_crossencoder=input_crossencoder,
        reranker=reranker,
        train_dataloader=test_dataloader,
        candidate_pool=candidate_pool,
        cand_encode_list=candidate_encoding,
        silent=params["silent"],
        logger=logger,
        top_k=params["top_k"],
        save_predictions=save_results,
    )

    if save_results:
        save_data_path = os.path.join(
            params["output_path"],
            "candidates_%s_top%d.t7" % (params["mode"], params["top_k"]),
        )
        torch.save(new_data, save_data_path)

    return new_data


# if __name__ == "__main__":
#     parser = BlinkParser(add_model_args=True)
#     parser.add_eval_args()
#     args = parser.parse_args()
#     print(args)
#     main(args.__dict__)



from bioel.models.arboel.biencoder.model.common.params import BlinkParser

import lightning as L
from pytorch_lightning.callbacks import ModelCheckpoint

from bioel.models.arboel.biencoder.model.BiEncoderLightningModule import LitArboel
from bioel.models.arboel.biencoder.data.BiEncoderLightningDataModule import (
    ArboelDataModule,
)
from pytorch_lightning.utilities import rank_zero_only
from bioel.ontology import BiomedicalOntology
import pickle
import os


@rank_zero_only
def create_ontology_object(args):
    if hasattr(BiomedicalOntology, args["load_function"]):
        load_func = getattr(BiomedicalOntology, args["load_function"])
        if args["ontology_dict"]:
            ontology_object = load_func(**args["ontology_dict"])
            print(f"Ontology loaded successfully. Name: {ontology_object.name}")
        else:
            raise ValueError("No ontology data provided.")
    else:
        raise ValueError(
            f"Error: {args['load_function']} is not a valid function for BiomedicalOntology."
        )

    file_path = os.path.join(args["data_path"], f"{args['ontology']}_object.pickle")
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, "wb") as f:
        pickle.dump(ontology_object, f)
    return ontology_object


def evaluate_model(params, model):

    create_ontology_object(params)
    os.makedirs(params["output_path"], exist_ok=True)

    data_module = ArboelDataModule(params=params)

    trainer = L.Trainer(
        limit_test_batches=1,
        devices=(
            params["devices"][:1]
            if isinstance(params["devices"], list)
            else params["devices"]
        ),
        accelerator="gpu",
        strategy="ddp_find_unused_parameters_true",
        enable_progress_bar=True,
        precision="16-mixed",
    )

    trainer.test(model=model, datamodule=data_module)



from typing import Optional, List, Dict, Tuple, NamedTuple, Type
import json
import datetime
from collections import defaultdict

import scipy
import numpy
import joblib
from sklearn.feature_extraction.text import TfidfVectorizer
from bioel.utils.file_cache import get_path

import nmslib
from nmslib.dist import FloatIndex
from bioel.models.scispacy.scispacy_embeddings import (
    KnowledgeBaseEmbeddings,
    load_approximate_nearest_neighbours_index,
    LinkerPaths,
)


class MentionCandidate(NamedTuple):
    """
    A data class representing a candidate entity that a mention may be linked to.

    Parameters
    ----------
    concept_id : str, required.
        The canonical concept id in the KB.
    aliases : List[str], required.
        The aliases that caused this entity to be linked.
    similarities : List[float], required.
        The cosine similarities from the mention text to the alias in tf-idf space.

    """

    concept_id: str
    aliases: List[str]
    similarities: List[float]


class CandidateGenerator:
    """
    A candidate generator for entity linking to a KnowledgeBase. Currently, two defaults are available:
     - Unified Medical Language System (UMLS).
     - Medical Subject Headings (MESH).

    To use these configured default KBs, pass the `name` parameter, either 'umls' or 'mesh'.

    It uses a sklearn.TfidfVectorizer to embed mention text into a sparse embedding of character 3-grams.
    These are then compared via cosine distance in a pre-indexed approximate nearest neighbours index of
    a subset of all entities and aliases in the KB.

    Once the K nearest neighbours have been retrieved, they are canonicalized to their KB canonical ids.
    This step is required because the index also includes entity aliases, which map to a particular canonical
    entity. This point is important for two reasons:

    1. K nearest neighbours will return a list of Y possible neighbours, where Y < K, because the entity ids
    are canonicalized.

    2. A single string may be an alias for multiple canonical entities. For example, "Jefferson County" may be an
    alias for both the canonical ids "Jefferson County, Iowa" and "Jefferson County, Texas". These are completely
    valid and important aliases to include, but it means that using the candidate generator to implement a naive
    k-nn baseline linker results in very poor performance, because there are multiple entities for some strings
    which have an exact char3-gram match, as these entities contain the same alias string. This situation results
    in multiple entities returned with a distance of 0.0, because they exactly match an alias, making a k-nn
    baseline effectively a random choice between these candidates. However, this doesn't matter if you have a
    classifier on top of the candidate generator, as is intended!

    Parameters
    ----------
    ann_index: FloatIndex
        An nmslib approximate nearest neighbours index.
    tfidf_vectorizer: TfidfVectorizer
        The vectorizer used to encode mentions.
    ann_concept_aliases_list: List[str]
        A list of strings, mapping the indices used in the ann_index to possible KB mentions.
        This is essentially used a lookup between the ann index and actual mention strings.
    kb: KnowledgeBase
        A class representing canonical concepts from the knowledge graph.
    verbose: bool
        Setting to true will print extra information about the generated candidates.
    ef_search: int
        The efs search parameter used in the index. This substantially effects runtime speed
        (higher is slower but slightly more accurate). Note that this parameter is ignored
        if a preconstructed ann_index is passed.
    name: str, optional (default = None)
        The name of the pretrained entity linker to load. Must be one of 'umls' or 'mesh'.
    """

    def __init__(
        self,
        serialized_kb: LinkerPaths,
        ann_index: Optional[FloatIndex] = None,
        tfidf_vectorizer: Optional[TfidfVectorizer] = None,
        ann_concept_aliases_list: Optional[List[str]] = None,
        verbose: bool = False,
        ef_search: int = 200,
        name: Optional[str] = None,
    ) -> None:
        """
        if name is not None and any(
            [ann_index, tfidf_vectorizer, ann_concept_aliases_list, kb]
        ):
            raise ValueError(
                "You cannot pass both a name argument and other constuctor arguments."
            )

        # Set the name to the default, after we have checked
        # the compatability with the args above.
        if name is None:
            name = "umls"

        linker_paths = DEFAULT_PATHS.get(name, UmlsLinkerPaths)
        """

        self.ann_index = ann_index or load_approximate_nearest_neighbours_index(
            serialized_kb=serialized_kb, ef_search=ef_search
        )
        self.vectorizer = tfidf_vectorizer or joblib.load(
            get_path(serialized_kb.tfidf_vectorizer)
        )
        self.ann_concept_aliases_list = ann_concept_aliases_list or json.load(
            open(get_path(serialized_kb.concept_aliases_list))
        )

        self.kb = serialized_kb

        self.verbose = verbose

    def nmslib_knn_with_zero_vectors(
        self, vectors: numpy.ndarray, k: int
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        ann_index.knnQueryBatch crashes if any of the vectors is all zeros.
        This function is a wrapper around `ann_index.knnQueryBatch` that solves this problem. It works as follows:
        - remove empty vectors from `vectors`.
        - call `ann_index.knnQueryBatch` with the non-empty vectors only. This returns `neighbors`,
        a list of list of neighbors. `len(neighbors)` equals the length of the non-empty vectors.
        - extend the list `neighbors` with `None`s in place of empty vectors.
        - return the extended list of neighbors and distances.
        """
        empty_vectors_boolean_flags = numpy.array(vectors.sum(axis=1) != 0).reshape(-1)
        empty_vectors_count = vectors.shape[0] - sum(empty_vectors_boolean_flags)
        if self.verbose:
            print(f"Number of empty vectors: {empty_vectors_count}")

        # init extended_neighbors with a list of Nones
        extended_neighbors = numpy.empty(
            (len(empty_vectors_boolean_flags),), dtype=object
        )
        extended_distances = numpy.empty(
            (len(empty_vectors_boolean_flags),), dtype=object
        )

        if vectors.shape[0] - empty_vectors_count == 0:
            return extended_neighbors, extended_distances

        # remove empty vectors before calling `ann_index.knnQueryBatch`
        vectors = vectors[empty_vectors_boolean_flags]

        # call `knnQueryBatch` to get neighbors
        original_neighbours = self.ann_index.knnQueryBatch(vectors, k=k)

        neighbors, distances = zip(
            *[(x[0].tolist(), x[1].tolist()) for x in original_neighbours]
        )
        neighbors = list(neighbors)
        distances = list(distances)

        # neighbors need to be converted to an np.array of objects instead of ndarray of dimensions len(vectors)xk
        # Solution: add a row to `neighbors` with any length other than k. This way, calling np.array(neighbors)
        # returns an np.array of objects
        neighbors.append([])
        distances.append([])
        # interleave `neighbors` and Nones in `extended_neighbors`
        extended_neighbors[empty_vectors_boolean_flags] = numpy.array(
            neighbors, dtype=object
        )[:-1]
        extended_distances[empty_vectors_boolean_flags] = numpy.array(
            distances, dtype=object
        )[:-1]

        return extended_neighbors, extended_distances

    def __call__(
        self, mention_texts: List[str], k: int
    ) -> List[List[MentionCandidate]]:
        """
        Given a list of mention texts, returns a list of candidate neighbors.

        NOTE: Because we include canonical name aliases in the ann index, the list
        of candidates returned will not necessarily be of length k for each candidate,
        because we then map these to canonical ids only.

        NOTE: For a given mention, the returned candidate list might be empty, which implies that
        the tfidf vector for this mention was all zeros (i.e there were no 3 gram overlaps). This
        happens reasonably rarely, but does occasionally.
        Parameters
        ----------
        mention_texts: List[str], required.
            The list of mention strings to generate candidates for.
        k: int, required.
            The number of ann neighbours to look up.
            Note that the number returned may differ due to aliases.

        Returns
        -------
        A list of MentionCandidate objects per mention containing KB concept_ids and aliases
        and distances which were mapped to. Note that these are lists for each concept id,
        because the index contains aliases which are canonicalized, so multiple values may map
        to the same canonical id.
        """

        if self.verbose:
            print(f"Generating candidates for {len(mention_texts)} mentions")

        # tfidf vectorizer crashes on an empty array, so we return early here
        if mention_texts == []:
            return []

        tfidfs = self.vectorizer.transform(mention_texts)
        start_time = datetime.datetime.now()

        # `ann_index.knnQueryBatch` crashes if one of the vectors is all zeros.
        # `nmslib_knn_with_zero_vectors` is a wrapper around `ann_index.knnQueryBatch` that addresses this issue.
        batch_neighbors, batch_distances = self.nmslib_knn_with_zero_vectors(tfidfs, k)
        end_time = datetime.datetime.now()
        total_time = end_time - start_time
        if self.verbose:
            print(f"Finding neighbors took {total_time.total_seconds()} seconds")
        batch_mention_candidates = []
        for neighbors, distances in zip(batch_neighbors, batch_distances):
            if neighbors is None:
                neighbors = []
            if distances is None:
                distances = []

            concept_to_mentions: Dict[str, List[str]] = defaultdict(list)
            concept_to_similarities: Dict[str, List[float]] = defaultdict(list)
            for neighbor_index, distance in zip(neighbors, distances):
                mention = self.ann_concept_aliases_list[neighbor_index]
                concepts_for_mention = self.kb.alias_to_cuis[mention]
                for concept_id in concepts_for_mention:
                    concept_to_mentions[concept_id].append(mention)
                    concept_to_similarities[concept_id].append(1.0 - distance)

            mention_candidates = [
                MentionCandidate(concept, mentions, concept_to_similarities[concept])
                for concept, mentions in concept_to_mentions.items()
            ]
            mention_candidates = sorted(mention_candidates, key=lambda c: c.concept_id)

            batch_mention_candidates.append(mention_candidates)

        return batch_mention_candidates



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
import os
import io
import copy
import torch
import logging
import numpy as np
import pickle
import json
from tqdm import tqdm, trange
from torch.utils.data import DataLoader, TensorDataset
from transformers import AutoTokenizer
import bioel.models.arboel.biencoder.data.data_process as data_process
from IPython import embed

from bioel.logger import setup_logger

import lightning as L
from bioel.models.arboel.biencoder.model.common.params import BlinkParser
from bioel.models.arboel.crossencoder.data.top_cands import save_topk_biencoder_cands
from bioel.models.arboel.biencoder.data.BiEncoderLightningDataModule import (
    read_dataset,
)
from bioel.models.arboel.biencoder.model.BiEncoderLightningModule import LitArboel
from datetime import datetime, timedelta
import torch.distributed as dist

# logger = setup_logger()
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

from pytorch_transformers.modeling_utils import CONFIG_NAME, WEIGHTS_NAME

OPTIM_SCHED_FNAME = "optim_sched.pth"


def modify(context_input, candidate_input, max_seq_length):
    new_input = []
    context_input = context_input.tolist()
    candidate_input = candidate_input.tolist()

    for i in tqdm(range(len(context_input)), desc="Concatenating"):
        cur_input = context_input[i]
        cur_candidate = candidate_input[i]
        mod_input = []
        for j in range(len(cur_candidate)):
            # remove [CLS] token from candidate
            sample = cur_input + cur_candidate[j][1:]
            sample = sample[:max_seq_length]
            mod_input.append(sample)

        new_input.append(mod_input)

    return torch.LongTensor(new_input)


def prepare_data(params):
    missing_files = any(
        not os.path.exists(
            os.path.join(
                params["biencoder_indices_path"],
                f"candidates_{data_split}_top64.t7",
            )
        )
        for data_split in ["train", "valid", "test"]
    )

    # Load the model and process data only if any file is missing
    if missing_files:
        device = torch.device(f"cuda:{torch.cuda.current_device()}")
        MyModel = LitArboel.load_from_checkpoint(
            checkpoint_path=params["biencoder_checkpoint"]
        )
        MyModel.to(device)  # Move to the current device
        reranker = MyModel.reranker

    for data_split in ["train", "valid", "test"]:
        fname = os.path.join(
            params["biencoder_indices_path"],
            f"candidates_{data_split}_top64.t7",
        )
        if not os.path.exists(fname):
            save_topk_biencoder_cands(
                bi_reranker=reranker,
                params=params,
                logger=logger,
                bi_tokenizer=AutoTokenizer.from_pretrained(
                    params["model_name_or_path"]
                ),
                topk=64,
            )


def load_data(
    data_split,
    bi_tokenizer,
    max_context_length,
    max_cand_length,
    knn,
    pickle_src_path,
    params,
    logger,
    return_dict_only=False,
):
    entity_dictionary_loaded = False
    entity_dictionary_pkl_path = os.path.join(
        pickle_src_path, "entity_dictionary.pickle"
    )
    if os.path.isfile(entity_dictionary_pkl_path):
        print("Loading stored processed entity dictionary...")
        with open(entity_dictionary_pkl_path, "rb") as read_handle:
            entity_dictionary = pickle.load(read_handle)
        entity_dictionary_loaded = True

    if return_dict_only and entity_dictionary_loaded:
        return entity_dictionary

    # Load data
    tensor_data_pkl_path = os.path.join(
        pickle_src_path, f"{data_split}_tensor_data.pickle"
    )
    processed_data_pkl_path = os.path.join(
        pickle_src_path, f"{data_split}_processed_data.pickle"
    )
    if os.path.isfile(tensor_data_pkl_path) and os.path.isfile(processed_data_pkl_path):
        print("Loading stored processed data...")
        with open(tensor_data_pkl_path, "rb") as read_handle:
            tensor_data = pickle.load(read_handle)
        with open(processed_data_pkl_path, "rb") as read_handle:
            processed_data = pickle.load(read_handle)
    else:
        data_samples = read_dataset(data_split, params["data_path"])
        if not entity_dictionary_loaded:
            with open(
                os.path.join(params["data_path"], "dictionary.pickle"), "rb"
            ) as read_handle:
                entity_dictionary = pickle.load(read_handle)

        # Check if dataset has multiple ground-truth labels
        mult_labels = "labels" in data_samples[0].keys()
        # Filter samples without gold entities
        data_samples = list(
            filter(
                lambda sample: (
                    (len(sample["labels"]) > 0)
                    if mult_labels
                    else (sample["label"] is not None)
                ),
                data_samples,
            )
        )
        logger.info("Read %d data samples." % len(data_samples))

        processed_data, entity_dictionary, tensor_data = (
            data_process.process_mention_data(
                data_samples,
                entity_dictionary,
                bi_tokenizer,
                max_context_length,
                max_cand_length,
                context_key=params["context_key"],
                multi_label_key="labels" if mult_labels else None,
                silent=params["silent"],
                logger=logger,
                debug=params["debug"],
                knn=knn,
                dictionary_processed=entity_dictionary_loaded,
            )
        )
        print("Saving processed data...")
        if not entity_dictionary_loaded:
            with open(entity_dictionary_pkl_path, "wb") as write_handle:
                pickle.dump(
                    entity_dictionary, write_handle, protocol=pickle.HIGHEST_PROTOCOL
                )
        with open(tensor_data_pkl_path, "wb") as write_handle:
            pickle.dump(tensor_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)
        with open(processed_data_pkl_path, "wb") as write_handle:
            pickle.dump(processed_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)

    if return_dict_only:
        return entity_dictionary
    return processed_data, entity_dictionary, tensor_data


def get_data_loader(
    params,
    data_split,
    tokenizer,
    context_length,
    candidate_length,
    max_seq_length,
    pickle_src_path,
    logger,
    inject_ground_truth=False,
    max_n=None,
    shuffle=True,
    return_data=False,
    custom_cand_set=None,
):
    # Load the top-64 indices for each mention query and the ground truth label if it exists in the candidate set
    logger.info(f"Loading {data_split} data...")
    cand_name = data_split
    if custom_cand_set is not None:
        logger.info(f"Loading custom candidate set: {custom_cand_set}...")
        cand_name = custom_cand_set
    fname = os.path.join(
        params["biencoder_indices_path"], f"candidates_{cand_name}_top64.t7"
    )
    stored_data = torch.load(fname)
    processed_data, entity_dictionary, tensor_data = load_data(
        data_split,
        tokenizer,
        context_length,
        candidate_length,
        1,
        pickle_src_path,
        params,
        logger,
    )
    logger.info("Loaded")
    dict_vecs = list(map(lambda x: x["ids"], entity_dictionary))

    mention_idxs = torch.tensor([i for i in range(len(stored_data["labels"]))])
    candidate_input = []
    keep_mask = [True] * len(stored_data["labels"])
    for i in tqdm(range(len(stored_data["labels"])), desc="Processing"):
        if stored_data["labels"][i] == -1:
            # If ground truth not in candidates, replace the last candidate with the ground truth
            if inject_ground_truth:
                gold_idx = processed_data[i]["label_idxs"][0]
                stored_data["labels"][i] = len(stored_data["candidates"][i]) - 1
                stored_data["candidates"][i][-1] = gold_idx
            else:  # skip the mention
                keep_mask[i] = False
                continue
        cands = list(map(lambda x: dict_vecs[x], stored_data["candidates"][i]))
        candidate_input.append(cands)
    candidate_input = np.array(candidate_input)
    context_input = tensor_data[:][0][keep_mask]
    label_input = torch.tensor(stored_data["labels"])[keep_mask]
    mention_idxs = mention_idxs[keep_mask]

    n_no_label = len(stored_data["labels"]) - np.sum(keep_mask)

    if max_n is not None:
        context_input = context_input[:max_n]
        candidate_input = candidate_input[:max_n]
        label_input = label_input[:max_n]

    if params["debug"]:
        max_n = 200
        context_input = context_input[:max_n]
        candidate_input = candidate_input[:max_n]
        label_input = label_input[:max_n]

    context_input = modify(context_input, candidate_input, max_seq_length)
    tensor_data = TensorDataset(context_input, label_input, mention_idxs)
    dataloader = DataLoader(
        tensor_data,
        shuffle=shuffle,
        num_workers=47,
        batch_size=params[
            "train_batch_size" if data_split == "train" else "eval_batch_size"
        ],
        drop_last=True if data_split == "train" else False,
    )
    if return_data:
        return (
            dataloader,
            n_no_label,
            {
                "entity_dictionary": entity_dictionary,
                "mention_data": processed_data,
                "stored_candidates": stored_data,
            },
        )
    return dataloader, n_no_label


class CrossEncoderDataModule(L.LightningDataModule):
    """ """

    def __init__(self, params):
        """
        Parameters
        ----------
        - params : dict(str)
        Contains configuration options
        """
        super(CrossEncoderDataModule, self).__init__()
        self.save_hyperparameters(params)

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.hparams["model_name_or_path"]
        )

    def prepare_data(self):
        pass

    def setup(self, stage=None):
        """
        For processing and splitting. Called at the beginning of fit (train + validate), test, or predict.
        """

        # Fit
        if stage == "fit" or stage is None:

            # Training set
            self.train_loader, _, self.train_data = get_data_loader(
                params=self.hparams,
                data_split="train",
                tokenizer=self.tokenizer,
                context_length=self.hparams["max_context_length"],
                candidate_length=self.hparams["max_context_length"],
                max_seq_length=self.hparams["max_seq_length"],
                pickle_src_path=self.hparams["data_path"],
                logger=logger,
                inject_ground_truth=self.hparams["inject_train_ground_truth"],
                return_data=True,
            )

            # Validation set
            self.valid_loader, _, self.valid_data = get_data_loader(
                params=self.hparams,
                data_split="valid",
                tokenizer=self.tokenizer,
                context_length=self.hparams["max_context_length"],
                candidate_length=self.hparams["max_context_length"],
                max_seq_length=self.hparams["max_seq_length"],
                pickle_src_path=self.hparams["data_path"],
                shuffle=False,
                logger=logger,
                inject_ground_truth=self.hparams["inject_eval_ground_truth"],
                return_data=True,
            )

        # Test
        if stage == "test":
            # Test set
            self.test_loader, self.nb_no_label_test, self.test_data = get_data_loader(
                params=self.hparams,
                data_split="test",
                tokenizer=self.tokenizer,
                context_length=self.hparams["max_context_length"],
                candidate_length=self.hparams["max_context_length"],
                max_seq_length=self.hparams["max_seq_length"],
                pickle_src_path=self.hparams["data_path"],
                shuffle=False,
                logger=logger,
                inject_ground_truth=self.hparams["inject_eval_ground_truth"],
                return_data=True,
            )

    def train_dataloader(self):
        # Return the training DataLoader
        return self.train_loader

    def val_dataloader(self):
        # Return the training DataLoader
        return self.valid_loader

    def test_dataloader(self):
        # Return the training DataLoader
        return self.test_loader



from bioel.models.arboel.biencoder.model.common.params import BlinkParser

import lightning as L
from pytorch_lightning.callbacks import ModelCheckpoint

from bioel.models.arboel.crossencoder.data.CrossEncoderLightningDataModule import (
    CrossEncoderDataModule,
)
from bioel.models.arboel.crossencoder.model.CrossEncoderLightningModule import (
    LitCrossEncoder,
)


def evaluate_model(params, model):
    data_module = CrossEncoderDataModule(params=params)

    trainer = L.Trainer(
        # limit_test_batches=5,
        limit_test_batches=(
            params["limit_test_batches"] if params["limit_test_batches"] else 1.0
        ),
        devices=params["devices"],
        accelerator="gpu",
        strategy="ddp_find_unused_parameters_true",
        precision="16-mixed",
    )

    trainer.test(model=model, datamodule=data_module)



##### Test file that you must put into the "package" folder #####
import unittest
from tqdm import tqdm

import spacy
from bioel.ontology import BiomedicalOntology
from bioel.models.scispacy.candidate_generation import CandidateGenerator
from bioel.models.scispacy.scispacy_embeddings import KnowledgeBaseEmbeddings
from bioel.models.scispacy.entity_linking import EntityLinker

class TestScispacy(unittest.TestCase):
    
	def test_kb_generation(self):
		test_cases = [
			{
				"file_path": "./kb_test2.jsonl",
				"name": "test_onto",
			},]

		for case in tqdm(test_cases):
			ontology = BiomedicalOntology.load_json(**case)
			myembed = KnowledgeBaseEmbeddings(ontology)                            				 # initializes the class for the embeddings given an Ontology in the memory
			myembed.create_tfidf_ann_index("./bioel/models/scispacy/kb_paths_scispacy")          # creates the tfidf vectors and the ann indexes
			kb = myembed.serialized_kb()                                         				 # returns a NamedTuple containing the necessariy paths to the embeddings and the needed dectionnaries in the memory
			print(kb)
   

	def test_candidate_generation(self):
		test_cases = [
		{
			"file_path": "./kb_test2.jsonl",
			"name": "test_onto",
		},]

		for case in tqdm(test_cases):
			ontology = BiomedicalOntology.load_json(**case)
			embeddings = KnowledgeBaseEmbeddings(ontology)                            				 # initializes the class for the embeddings given an Ontology in the memory
			embeddings.create_tfidf_ann_index("./bioel/models/scispacy/kb_paths_scispacy")           # creates the tfidf vectors and the ann indexes
			kb = embeddings.serialized_kb()                                         				 # returns a NamedTuple containing the necessariy paths to the embeddings and the needed dectionnaries in the memory
			
			generator = CandidateGenerator(kb)         # initializes the candidate generator for the serialized KB                                 
			mentions = ["enzyme"]                      # simple example with 1 mention 
			k = 10                                     # number of K nearest neighbors
			candidates = generator(mentions,k)         # calls the candidate generator for this mention
			print(candidates)                          # We display all the top candidates

	def test_entity_linker(self):

		test_cases = [
			{
				"file_path": "./kb_test2.jsonl",
				"name": "test_onto",
			},]

		for case in tqdm(test_cases):
			ontology = BiomedicalOntology.load_json(**case)
			embeddings = KnowledgeBaseEmbeddings(ontology)                            				 # initializes the class for the embeddings given an Ontology in the memory
			embeddings.create_tfidf_ann_index("./bioel/models/scispacy/kb_paths_scispacy")           # creates the tfidf vectors and the ann indexes
			kb = embeddings.serialized_kb()                                         				 # returns a NamedTuple containing the necessariy paths to the embeddings and the needed dectionnaries in the memory

			nlp = spacy.load("en_core_sci_sm")               # Load a Pretrained model "en_core_sci_sm" available at https://github.com/allenai/scispacy/tree/main
															 # this is the small model but one can feel free to use other available models or train its own model 
															 # by using Spacy's specific config files 
			linker = EntityLinker(kb, threshold = 0.6)       # adjusts the similarity score threshold if you don't find any entity for a mention

			doc = nlp("Branching Enzyme")                    # simple example with a 2 words text
			print(doc.ents)

			linker(doc)
			entity = doc.ents[1]           # we pick the second mention token "Enzyme"
			print(entity)                  # returns a tuple containing the identified mentions by the NER
			print(entity._.kb_ents)        # returns a list of tuples containing the cuis and the matching score linked to each mention in the text

			for umls_ent in entity._.kb_ents:
				print(linker.kb.cui_to_entity[umls_ent[0]])           # retrieves all the information about the entity as defined in BiomedicalEntity class given the cui in umls_ent[0]

if __name__ == "__main__":
    unittest.main()


import numpy as np
import pandas as pd
import seaborn as sns
import sys
import time
import itertools
import pickle

from tqdm.auto import tqdm
from collections import defaultdict
from matplotlib import pyplot as plt

from bigbio_utils import (
    dataset_to_df,
    DATASET_NAMES,
    CUIS_TO_EXCLUDE,
    CUIS_TO_REMAP,
    resolve_abbreviation,
    dataset_to_documents,
    load_bigbio_dataset,
)
from dataset_consts import dataset_to_pretty_name, model_to_pretty_name, model_to_color

sns.set()
sns.set_style("whitegrid")

models = [
    "sapbert",
    "metamap",
    "krissbert",
    "scispacy",
    "medlinker",
    "arboel_biencoder",
    "arboel_crossencoder",
    "biobart",
    "biogenel",
]

to_add = ["bern2", "bootleg"]

pd.set_option("display.max_columns", 80)


def metamap_to_candidates(metamap_output, mappings_first=False):
    if mappings_first:
        cols = ["text", "mapping_cui_list", "candidate_cui_list"]
    else:
        cols = [
            "text",
            "candidate_cui_list",
            "mapping_cui_list",
        ]

    # Create mapping from text to candidates for MedMentions
    text2candidates = defaultdict(list)

    for row in metamap_output[cols].values:
        text = row[0]
        candidates = eval(row[1])
        for c in eval(row[2]):
            candidates
        # TODO: Need to account for correct database for
        candidates = [
            ["UMLS:" + x if ":" not in x else x.replace("ncbigene", "NCBIGene")]
            for x in candidates
        ]
        text2candidates[text] = candidates

    return text2candidates


def deduplicate_candidates(candidates, return_counts=False):
    deduplicated = []
    counts = defaultdict(int)
    for c in candidates:
        if type(c) == list:
            deduped_subset = [x for x in c if x not in counts]
            if len(deduped_subset) > 0:
                deduplicated.append(deduped_subset)

            for x in c:
                counts[x] += 1

        else:
            counts[c] += 1
            if c not in deduplicated:
                deduplicated.append(c)

    if return_counts:
        return deduplicated, counts

    return deduplicated


def output_list_to_candidates(output_list, k=20, sort_by_counts=False):
    """
    Turn model output into candidate list
    """
    output2candidates = defaultdict(list)
    for mention in output_list:
        doc_id = mention["document_id"]
        offsets = mention["offsets"]
        joined_offsets = ";".join([",".join([str(y) for y in x]) for x in offsets])
        # candidates = [y for x in mention['candidates'] for y in x]
        candidates = mention["candidates"]
        deduplicated_candidates, counts = deduplicate_candidates(
            candidates, return_counts=True
        )

        # Syntax to prioritize candidates that come up more frequently in retrieved results
        if sort_by_counts:
            top_cuis = sorted(
                [(k, v) for k, v in counts.items()], key=lambda x: x[1], reverse=True
            )
            grouped_candidates = []
            for k, g in itertools.groupby(top_cuis, key=lambda x: x[1]):
                grouped_candidates.append(list(g))

            output2candidates[(doc_id, joined_offsets)] = grouped_candidates
        else:
            output2candidates[(doc_id, joined_offsets)] = deduplicated_candidates[:k]

    return output2candidates


def add_candidates_to_df(df, candidate_dict, new_col_name, eval_col="text"):
    """
    Add candidates produced by a model to results dataframe for evaluation

    Parameters:
    ----------------------------
        df: pandas.DataFrame
            Dataframe of bigbio mention produced by bigbio_utils.dataset_to_df

        candidate_dict: dict
            Mapping of (document_id, joined_offsets) -> candidates

        new_col_name: str
            Name of new column

        eval_col: Column to evaluate on
    """
    df[new_col_name] = df[["document_id", "joined_offsets"]].apply(
        lambda x: candidate_dict[(x[0], x[1])], axis=1
    )


def list_flatten(nested_list):
    used = set([])
    flattened = []
    for x in nested_list:
        for y in x:
            if y not in used:
                flattened.append(y)
                used.add(y)

    return flattened


def min_hit_index(gold_cuis, candidates, eval_mode):
    """
    Find index of first hit in candidates
    """
    if eval_mode == "basic":
        flat_candidates = list_flatten(candidates)
        for i, c in enumerate(flat_candidates):
            if c in gold_cuis:
                return i
    elif eval_mode == "strict":
        for i, c in enumerate(candidates):
            if all(x in gold_cuis for x in c):
                return i

    elif eval_mode == "relaxed":
        for i, c in enumerate(candidates):
            if any(x in gold_cuis for x in c):
                return i

    else:
        raise ValueError(f"eval_mode {eval_mode} not supported")

    return 1000000


def recall_at_k(
    df,
    candidate_col,
    gold_col="db_ids",
    max_k: int = 10,
    filter_null=False,
    eval_mode="basic",
):
    """
    Compute recall@k for all values of k < max_k
    """
    # Filter rows wilt null values after CUI remapping
    if filter_null:
        before_row_count = df.shape[0]
        df = df[df[gold_col].map(lambda x: len(x) > 0)]
        after_row_count = df.shape[0]
        print(f"Dropped {before_row_count - after_row_count} rows with null db_ids")

    df[f"{candidate_col}_min_hit_index"] = df[[gold_col, candidate_col]].apply(
        lambda x: min_hit_index(x[0], x[1], eval_mode=eval_mode), axis=1
    )

    recall_at_k_dict = {}
    for k in range(1, max_k + 1):
        recall_at_k_dict[k] = (df[f"{candidate_col}_min_hit_index"] < k).mean()

    return recall_at_k_dict


def plot_recall_at_k(
    recall_dict, max_k=10, legend_key=None, ax=None, color=None, alpha=1
):
    if ax is not None:
        ax.plot(
            recall_dict.keys(),
            recall_dict.values(),
            marker="o",
            label=legend_key,
            color=color,
            alpha=alpha,
        )
    else:
        plt.plot(
            recall_dict.keys(),
            recall_dict.values(),
            marker="o",
            label=legend_key,
            color=color,
            alpha=alpha,
        )
    # if legend_key is not None:
    #     plt.legend()


def correct_medlinker_candidates(medlinker_output):
    for x in medlinker_output:
        new_candidates = []
        for y in x["candidates"]:
            single_candidate = []
            for c in y:
                if c is None:
                    continue
                if not c.startswith("MESH") and not c.startswith("OMIM"):
                    c = c.replace("ESH", "MESH")
                single_candidate.append(c)
            new_candidates.append(single_candidate)
        x["candidates"] = new_candidates
    return medlinker_output



from spacy.tokens import Doc
from spacy.tokens import Span
from spacy.language import Language

from bioel.models.scispacy.candidate_generation import CandidateGenerator
from bioel.models.scispacy.scispacy_embeddings import LinkerPaths
from typing import Optional


class EntityLinker:
    """
    A spacy pipeline component which identifies entities in text which appear
    in a knowledge base.

    Currently, there are five defaults: the Unified Medical Language System (UMLS),
    the Medical Subject Headings (MeSH) dictionary, the RxNorm ontology, the Gene
    Ontology, and the Human Phenotype Ontology.

    To use these configured default KBs, pass the `name` parameter ('umls','mesh',
    'rxnorm','go','hpo').

    Currently this implementation just compares string similarity, returning
    entities above a given threshold.

    This class sets the `._.kb_ents` attribute on spacy Spans, which consists of a
    List[Tuple[str, float]] corresponding to the KB concept_id and the associated score
    for a list of `max_entities_per_mention` number of entities.

    You can look up more information for a given id using the kb attribute of this class:

    print(linker.kb.cui_to_entity[concept_id])

    A Note on Definitions:
    Only 187767 entities, or 6.74% of the UMLS KB have definitions. However,
    the MedMentions dataset links to entities which have definitions 82.9% of the time. So by
    default, we only link to entities which have definitions (typically they are more salient / cleaner),
    but this might not suit your use case. YMMV.


    Parameters
    ----------

    nlp: `Language`, a required argument for spacy to use this as a factory
    name: `str`, a required argument for spacy to use this as a factory
    candidate_generator : `CandidateGenerator`, optional, (default = None)
        A CandidateGenerator to generate entity candidates for mentions.
        If no candidate generator is passed, the default pretrained one is used.
    resolve_abbreviations : bool = True, optional (default = False)
        Whether to resolve abbreviations identified in the Doc before performing linking.
        This parameter has no effect if there is no `AbbreviationDetector` in the spacy
        pipeline.
    k : int, optional, (default = 30)
        The number of nearest neighbours to look up from the candidate generator per mention.
    threshold : float, optional, (default = 0.7)
        The threshold that a entity candidate must reach to be added to the mention in the Doc
        as a mention candidate.
    no_definition_threshold : float, optional, (default = 0.95)
        The threshold that a entity candidate must reach to be added to the mention in the Doc
        as a mention candidate if the entity candidate does not have a definition.
    filter_for_definitions: bool, default = True
        Whether to filter entities that can be returned to only include those with definitions
        in the knowledge base.
    max_entities_per_mention : int, optional, default = 5
        The maximum number of entities which will be returned for a given mention, regardless of
        how many are nearest neighbours are found.
    linker_name: str, optional (default = None)
        The name of the pretrained entity linker to load.
    """

    def __init__(
        self,
        serialized_kb: LinkerPaths,
        nlp: Optional[Language] = None,
        name: str = "scispacy_linker",
        candidate_generator: Optional[CandidateGenerator] = None,
        resolve_abbreviations: bool = True,
        k: int = 30,
        threshold: float = 0.7,
        no_definition_threshold: float = 0.95,
        filter_for_definitions: bool = True,
        max_entities_per_mention: int = 5,
    ):
        Span.set_extension("kb_ents", default=[], force=True)

        self.candidate_generator = candidate_generator or CandidateGenerator(
            serialized_kb=serialized_kb
        )
        self.resolve_abbreviations = resolve_abbreviations
        self.k = k
        self.threshold = threshold
        self.no_definition_threshold = no_definition_threshold
        self.kb = serialized_kb
        self.filter_for_definitions = filter_for_definitions
        self.max_entities_per_mention = max_entities_per_mention

    def __call__(self, doc: Doc) -> Doc:
        mention_strings = []
        if self.resolve_abbreviations and Doc.has_extension("abbreviations"):
            # TODO: This is possibly sub-optimal - we might
            # prefer to look up both the long and short forms.
            for ent in doc.ents:
                if isinstance(ent._.long_form, Span):
                    # Long form
                    mention_strings.append(ent._.long_form.text)
                elif isinstance(ent._.long_form, str):
                    # Long form
                    mention_strings.append(ent._.long_form)
                else:
                    # no abbreviations case
                    mention_strings.append(ent.text)
        else:
            mention_strings = [x.text for x in doc.ents]

        batch_candidates = self.candidate_generator(mention_strings, self.k)

        for mention, candidates in zip(doc.ents, batch_candidates):
            predicted = []
            for cand in candidates:
                score = max(cand.similarities)
                if (
                    self.filter_for_definitions
                    and self.kb.cui_to_entity[cand.concept_id].definition is None
                    and score < self.no_definition_threshold
                ):
                    continue
                if score > self.threshold:
                    predicted.append((cand.concept_id, score))
            sorted_predicted = sorted(predicted, reverse=True, key=lambda x: x[1])
            mention._.kb_ents = sorted_predicted[: self.max_entities_per_mention]

        return doc



# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) 2021 Dhruv Agarwal and authors of arboEL.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
import collections
import os
import sys
import io
import copy
import torch
import logging
import random
import time
import numpy as np
import pickle
import json
from tqdm import tqdm, trange
from torch.utils.data import DataLoader, TensorDataset
from transformers import AutoTokenizer
from bioel.models.arboel.model.common.optimizer import get_bert_optimizer
from pytorch_transformers.optimization import WarmupLinearSchedule
from bioel.models.arboel.crossencoder.original.crossencoder import CrossEncoderRanker
import bioel.models.arboel.data.data_process as data_process
from IPython import embed

from bioel.logger import setup_logger

from lightning.pytorch.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import WandbLogger
import lightning as L
from datetime import datetime
from bioel.models.arboel.model.common.params import BlinkParser

# logger = setup_logger()
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

from pytorch_transformers.modeling_utils import CONFIG_NAME, WEIGHTS_NAME

OPTIM_SCHED_FNAME = "optim_sched.pth"


def read_dataset(dataset_name, preprocessed_json_data_parent_folder, debug=False):
    file_name = "{}.jsonl".format(dataset_name)
    txt_file_path = os.path.join(preprocessed_json_data_parent_folder, file_name)

    samples = []

    with io.open(txt_file_path, mode="r", encoding="utf-8") as file:
        for line in file:
            samples.append(json.loads(line.strip()))
            if debug and len(samples) > 200:
                break

    print(f"Read {len(samples)} samples.")
    return samples


def accuracy(out, labels, return_bool_arr=False):
    outputs = np.argmax(out, axis=1)
    if return_bool_arr:
        return outputs == labels, outputs
    return np.sum(outputs == labels)


def modify(context_input, candidate_input, max_seq_length):
    new_input = []
    context_input = context_input.tolist()
    candidate_input = candidate_input.tolist()

    for i in tqdm(range(len(context_input)), desc="Concatenating"):
        cur_input = context_input[i]
        cur_candidate = candidate_input[i]
        mod_input = []
        for j in range(len(cur_candidate)):
            # remove [CLS] token from candidate
            sample = cur_input + cur_candidate[j][1:]
            sample = sample[:max_seq_length]
            mod_input.append(sample)

        new_input.append(mod_input)

    return torch.LongTensor(new_input)


def convert_defaultdict(d):
    if isinstance(d, collections.defaultdict):
        d = {k: convert_defaultdict(v) for k, v in d.items()}
    elif isinstance(d, dict):
        return {k: convert_defaultdict(v) for k, v in d.items()}
    elif isinstance(d, np.int64):
        return int(d)
    elif isinstance(d, np.float64):
        return float(d)
    return d


def merge_dicts(dict1, dict2):
    """Merge two dictionaries with support for nested structures and summing numeric values"""
    merged_dict = dict1.copy()
    for key, value in dict2.items():
        if key in merged_dict:
            if isinstance(value, collections.defaultdict) and isinstance(
                merged_dict[key], collections.defaultdict
            ):
                for k, v in value.items():
                    merged_dict[key][k] = merged_dict[key].get(k, 0) + v
            elif isinstance(value, (int, float)) and isinstance(
                merged_dict[key], (int, float)
            ):
                merged_dict[key] += value
            else:
                if isinstance(value, list) and isinstance(merged_dict[key], list):
                    merged_dict[key].extend(value)
                else:
                    merged_dict[key] = value
        else:
            merged_dict[key] = copy.deepcopy(value)

    return merged_dict


def evaluate_single_batch(
    reranker,
    batch,
    logger,
    context_length,
    mention_data=None,
    compute_macro_avg=False,
    store_failure_success=False,
):
    context_input, label_input, mention_idxs = batch
    results = {}
    eval_accuracy = 0
    nb_eval_examples = 0

    if mention_data is not None:
        processed_mention_data = mention_data["mention_data"]
        dictionary = mention_data["entity_dictionary"]
        stored_candidates = mention_data["stored_candidates"]
        if store_failure_success:
            failsucc = {"failure": [], "success": []}

    n_evaluated_per_type = collections.defaultdict(int)
    n_hits_per_type = collections.defaultdict(int)

    with torch.no_grad():
        eval_loss, logits = reranker(context_input, label_input, context_length)

    logits = logits.detach().cpu().numpy()
    label_ids = label_input.cpu().numpy()

    tmp_eval_hits, predicted = accuracy(logits, label_ids, return_bool_arr=True)
    tmp_eval_accuracy = np.sum(tmp_eval_hits)

    eval_accuracy = tmp_eval_accuracy
    nb_eval_examples = context_input.size(0)

    if compute_macro_avg:
        for i, m_idx in enumerate(mention_idxs):
            mention_type = processed_mention_data[m_idx]["type"]
            n_evaluated_per_type[mention_type] += 1
            is_hit = tmp_eval_hits[i]
            n_hits_per_type[mention_type] += is_hit

    if store_failure_success:
        for i, m_idx in enumerate(mention_idxs):
            m_idx = m_idx.item()
            men_query = processed_mention_data[m_idx]
            dict_pred = dictionary[stored_candidates["candidates"][m_idx][predicted[i]]]
            report_obj = {
                "mention_id": men_query["mention_id"],
                "mention_name": men_query["mention_name"],
                "mention_gold_cui": "|".join(men_query["label_cuis"]),
                "mention_gold_cui_name": "|".join(
                    [
                        dictionary[i]["title"]
                        for i in men_query["label_idxs"][: men_query["n_labels"]]
                    ]
                ),
                "predicted_name": dict_pred["title"],
                "predicted_cui": dict_pred["cui"],
            }
            failsucc["success" if tmp_eval_hits[i] else "failure"].append(report_obj)

    results["nb_samples_evaluated"] = nb_eval_examples
    results["correct_pred"] = int(eval_accuracy)

    print(f"Eval: Accuracy: {eval_accuracy/nb_eval_examples*100}%")

    if compute_macro_avg:
        results["n_hits_per_type"] = n_hits_per_type
        results["n_evaluated_per_type"] = n_evaluated_per_type

    if store_failure_success:
        results["failure"] = failsucc["failure"]
        results["success"] = failsucc["success"]

    if not store_failure_success:
        logger.info(json.dumps(results))

    return results


def load_data(
    data_split,
    bi_tokenizer,
    max_context_length,
    max_cand_length,
    knn,
    pickle_src_path,
    params,
    logger,
    return_dict_only=False,
):
    entity_dictionary_loaded = False
    entity_dictionary_pkl_path = os.path.join(
        pickle_src_path, "entity_dictionary.pickle"
    )
    if os.path.isfile(entity_dictionary_pkl_path):
        print("Loading stored processed entity dictionary...")
        with open(entity_dictionary_pkl_path, "rb") as read_handle:
            entity_dictionary = pickle.load(read_handle)
        entity_dictionary_loaded = True

    if return_dict_only and entity_dictionary_loaded:
        return entity_dictionary

    # Load data
    tensor_data_pkl_path = os.path.join(
        pickle_src_path, f"{data_split}_tensor_data.pickle"
    )
    processed_data_pkl_path = os.path.join(
        pickle_src_path, f"{data_split}_processed_data.pickle"
    )
    if os.path.isfile(tensor_data_pkl_path) and os.path.isfile(processed_data_pkl_path):
        print("Loading stored processed data...")
        with open(tensor_data_pkl_path, "rb") as read_handle:
            tensor_data = pickle.load(read_handle)
        with open(processed_data_pkl_path, "rb") as read_handle:
            processed_data = pickle.load(read_handle)
    else:
        data_samples = read_dataset(data_split, params["data_path"])
        if not entity_dictionary_loaded:
            with open(
                os.path.join(params["data_path"], "dictionary.pickle"), "rb"
            ) as read_handle:
                entity_dictionary = pickle.load(read_handle)

        # Check if dataset has multiple ground-truth labels
        mult_labels = "labels" in data_samples[0].keys()
        # Filter samples without gold entities
        data_samples = list(
            filter(
                lambda sample: (
                    (len(sample["labels"]) > 0)
                    if mult_labels
                    else (sample["label"] is not None)
                ),
                data_samples,
            )
        )
        logger.info("Read %d data samples." % len(data_samples))

        processed_data, entity_dictionary, tensor_data = (
            data_process.process_mention_data(
                data_samples,
                entity_dictionary,
                bi_tokenizer,
                max_context_length,
                max_cand_length,
                context_key=params["context_key"],
                multi_label_key="labels" if mult_labels else None,
                silent=params["silent"],
                logger=logger,
                debug=params["debug"],
                knn=knn,
                dictionary_processed=entity_dictionary_loaded,
            )
        )
        print("Saving processed data...")
        if not entity_dictionary_loaded:
            with open(entity_dictionary_pkl_path, "wb") as write_handle:
                pickle.dump(
                    entity_dictionary, write_handle, protocol=pickle.HIGHEST_PROTOCOL
                )
        with open(tensor_data_pkl_path, "wb") as write_handle:
            pickle.dump(tensor_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)
        with open(processed_data_pkl_path, "wb") as write_handle:
            pickle.dump(processed_data, write_handle, protocol=pickle.HIGHEST_PROTOCOL)

    if return_dict_only:
        return entity_dictionary
    return processed_data, entity_dictionary, tensor_data


def get_data_loader(
    params,
    data_split,
    tokenizer,
    context_length,
    candidate_length,
    max_seq_length,
    pickle_src_path,
    logger,
    inject_ground_truth=False,
    max_n=None,
    shuffle=True,
    return_data=False,
    custom_cand_set=None,
):
    # Load the top-64 indices for each mention query and the ground truth label if it exists in the candidate set
    logger.info(f"Loading {data_split} data...")
    cand_name = data_split
    if custom_cand_set is not None:
        logger.info(f"Loading custom candidate set: {custom_cand_set}...")
        cand_name = custom_cand_set
    fname = os.path.join(
        params["biencoder_indices_path"], f"candidates_{cand_name}_top64.t7"
    )
    stored_data = torch.load(fname)
    processed_data, entity_dictionary, tensor_data = load_data(
        data_split,
        tokenizer,
        context_length,
        candidate_length,
        1,
        pickle_src_path,
        params,
        logger,
    )
    logger.info("Loaded")
    dict_vecs = list(map(lambda x: x["ids"], entity_dictionary))

    mention_idxs = torch.tensor([i for i in range(len(stored_data["labels"]))])
    candidate_input = []
    keep_mask = [True] * len(stored_data["labels"])
    for i in tqdm(range(len(stored_data["labels"])), desc="Processing"):
        if stored_data["labels"][i] == -1:
            # If ground truth not in candidates, replace the last candidate with the ground truth
            if inject_ground_truth:
                gold_idx = processed_data[i]["label_idxs"][0]
                stored_data["labels"][i] = len(stored_data["candidates"][i]) - 1
                stored_data["candidates"][i][-1] = gold_idx
            else:
                keep_mask[i] = False
                continue
        cands = list(map(lambda x: dict_vecs[x], stored_data["candidates"][i]))
        candidate_input.append(cands)
    candidate_input = np.array(candidate_input)
    context_input = tensor_data[:][0][keep_mask]
    label_input = torch.tensor(stored_data["labels"])[keep_mask]
    mention_idxs = mention_idxs[keep_mask]

    n_no_label = len(stored_data["labels"]) - np.sum(keep_mask)

    if max_n is not None:
        context_input = context_input[:max_n]
        candidate_input = candidate_input[:max_n]
        label_input = label_input[:max_n]

    if params["debug"]:
        max_n = 200
        context_input = context_input[:max_n]
        candidate_input = candidate_input[:max_n]
        label_input = label_input[:max_n]

    context_input = modify(context_input, candidate_input, max_seq_length)
    tensor_data = TensorDataset(context_input, label_input, mention_idxs)
    dataloader = DataLoader(
        tensor_data,
        shuffle=shuffle,
        batch_size=params[
            "train_batch_size" if data_split == "train" else "eval_batch_size"
        ],
    )
    if return_data:
        return (
            dataloader,
            n_no_label,
            {
                "entity_dictionary": entity_dictionary,
                "mention_data": processed_data,
                "stored_candidates": stored_data,
            },
        )
    return dataloader, n_no_label


class CrossEncoderDataModule(L.LightningDataModule):
    """ """

    def __init__(self, params):
        """
        Parameters
        ----------
        - params : dict(str)
        Contains configuration options
        """
        super(CrossEncoderDataModule, self).__init__()
        self.save_hyperparameters(params)

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.hparams["model_name_or_path"]
        )

    def setup(self, stage=None):
        """
        For processing and splitting. Called at the beginning of fit (train + validate), test, or predict.
        """

        # Fit
        if stage == "fit" or stage is None:

            # Training set
            self.train_loader, _, self.train_data = get_data_loader(
                params=self.hparams,
                data_split="train",
                tokenizer=self.tokenizer,
                context_length=self.hparams["max_context_length"],
                candidate_length=self.hparams["max_context_length"],
                max_seq_length=self.hparams["max_seq_length"],
                pickle_src_path=self.hparams["data_path"],
                logger=logger,
                inject_ground_truth=self.hparams["inject_train_ground_truth"],
                return_data=True,
            )

            # Validation set
            self.valid_loader, _, self.valid_data = get_data_loader(
                params=self.hparams,
                data_split="valid",
                tokenizer=self.tokenizer,
                context_length=self.hparams["max_context_length"],
                candidate_length=self.hparams["max_context_length"],
                max_seq_length=self.hparams["max_seq_length"],
                pickle_src_path=self.hparams["data_path"],
                logger=logger,
                inject_ground_truth=self.hparams["inject_eval_ground_truth"],
                return_data=True,
            )

        # Test
        if stage == "test":
            # Test set
            self.test_loader, self.nb_no_label_test, self.test_data = get_data_loader(
                params=self.hparams,
                data_split="test",
                tokenizer=self.tokenizer,
                context_length=self.hparams["max_context_length"],
                candidate_length=self.hparams["max_context_length"],
                max_seq_length=self.hparams["max_seq_length"],
                pickle_src_path=self.hparams["data_path"],
                logger=logger,
                inject_ground_truth=self.hparams["inject_eval_ground_truth"],
                return_data=True,
            )

    def train_dataloader(self):
        # Return the training DataLoader
        return self.train_loader

    def val_dataloader(self):
        # Return the training DataLoader
        return self.valid_loader

    def test_dataloader(self):
        # Return the training DataLoader
        return self.test_loader


class CrossEncoder(L.LightningModule):
    """ """

    def __init__(self, params):
        super(CrossEncoder, self).__init__()
        self.save_hyperparameters(params)
        self.reranker = CrossEncoderRanker(params)
        self.model = self.reranker.model
        self.val_results = {}

    def training_step(self, batch, batch_idx):
        context_input, label_input, _ = batch
        train_loss, _ = self.reranker(
            context_input, label_input, self.hparams["max_context_length"]
        )

        print("training loss :", train_loss)
        self.log(
            "train loss :", train_loss, on_step=True, on_epoch=True, sync_dist=True
        )

        return train_loss

    def configure_optimizers(self):
        # Define optimizer
        optimizer = get_bert_optimizer(
            models=[self.model],
            type_optimization=self.hparams["type_optimization"],
            learning_rate=self.hparams["learning_rate"],
            fp16=self.hparams.get("fp16"),
        )

        # Define scheduler
        num_train_steps = (
            int(
                len(self.trainer.datamodule.train_data["mention_data"])
                / self.hparams["train_batch_size"]
                / self.trainer.accumulate_grad_batches
            )
            * self.trainer.max_epochs
        )
        num_warmup_steps = int(num_train_steps * self.hparams["warmup_proportion"])

        scheduler = WarmupLinearSchedule(
            optimizer=optimizer,
            warmup_steps=num_warmup_steps,
            t_total=num_train_steps,
        )
        logger.info(" Num optimization steps = %d" % num_train_steps)
        logger.info(" Num warmup steps = %d", num_warmup_steps)
        return [optimizer], [{"scheduler": scheduler, "interval": "step"}]

    def validation_step(self, batch, batch_idx):

        results = evaluate_single_batch(
            reranker=self.reranker,
            batch=batch,
            logger=logger,
            context_length=self.hparams["max_context_length"],
        )

        self.log(
            "max_acc",
            results["normalized_accuracy"],
            on_epoch=True,
            sync_dist=True,
        )

    def on_test_epoch_start(self):

        self.n_evaluated_per_type = {}
        self.n_mentions_per_type = {}

        self.test_results = {
            "normalized_accuracy": 0,
            "unnormalized_accuracy": 0,
            "normalized_macro_avg_acc": 0,
            "unnormalized_macro_avg_acc": 0,
            "filtered_length": int(
                len(self.trainer.datamodule.test_data["mention_data"])
                - self.trainer.datamodule.nb_no_label_test
            ),
            "unfiltered_length": len(self.trainer.datamodule.test_data["mention_data"]),
            "n_mentions_per_type": collections.defaultdict(int),
            "n_hits_per_type": collections.defaultdict(int),
            "n_evaluated_per_type": collections.defaultdict(int),
        }

    def test_step(self, batch, batch_idx):

        results = evaluate_single_batch(
            reranker=self.reranker,
            batch=batch,
            logger=logger,
            context_length=self.hparams["max_context_length"],
            mention_data=self.trainer.datamodule.test_data,
            compute_macro_avg=True,
            store_failure_success=True,
        )
        self.test_results = merge_dicts(self.test_results, results)

    def on_test_epoch_end(self):

        self.test_results["normalized_accuracy"] = float(
            self.test_results["correct_pred"]
            / self.test_results["nb_samples_evaluated"]
            * 100
        )

        self.test_results["unnormalized_accuracy"] = float(
            self.test_results["normalized_accuracy"]
            * self.test_results["filtered_length"]
            / self.test_results["unfiltered_length"]
        )

        for mention_type in self.test_results["n_evaluated_per_type"]:
            self.test_results["normalized_macro_avg_acc"] += float(
                self.test_results["n_hits_per_type"][mention_type]
                / self.test_results["n_evaluated_per_type"][mention_type]
                if self.test_results["n_evaluated_per_type"][mention_type] > 0
                else 0
            )

        self.test_results["normalized_macro_avg_acc"] = (
            self.test_results["normalized_macro_avg_acc"]
            / len(self.test_results["n_evaluated_per_type"])
            * 100
        )

        self.log(
            "Number of samples evaluated in this test",
            self.test_results["nb_samples_evaluated"],
            sync_dist=True,
        )

        self.log(
            "Normalized Accuracy : Only considered mentions with the correct cui in top candidates",
            self.test_results["normalized_accuracy"],
            sync_dist=True,
        )

        self.log(
            "Number of samples with the correct cui in top candidates in the test set",
            self.test_results["filtered_length"],
            sync_dist=True,
        )
        self.log(
            "Number of samples with and without the correct cui in top candidatesin the test set",
            self.test_results["unfiltered_length"],
            sync_dist=True,
        )

        self.log(
            "Unnormalized Accuracy : Include mentions without the correct cui in top candidates",
            self.test_results["unnormalized_accuracy"],
            sync_dist=True,
        )
        self.log(
            "Normalized macro_avg_acc : Average accuracy across all types (Only mentions with the correct cui in top candidates).",
            self.test_results["normalized_macro_avg_acc"],
            sync_dist=True,
        )

        if (
            self.test_results["filtered_length"]
            == self.test_results["nb_samples_evaluated"]
        ):
            for men in self.trainer.datamodule.test_data["mention_data"]:
                self.test_results["n_mentions_per_type"][men["type"]] += 1

            for mention_type in self.test_results["n_mentions_per_type"]:
                self.test_results["unnormalized_macro_avg_acc"] += float(
                    self.test_results["n_hits_per_type"][mention_type]
                    / self.test_results["n_mentions_per_type"][mention_type]
                    if self.test_results["n_mentions_per_type"][mention_type] > 0
                    else 0
                )

            self.test_results["unnormalized_macro_avg_acc"] = (
                self.test_results["unnormalized_macro_avg_acc"]
                / len(self.test_results["n_mentions_per_type"])
                * 100
            )
            self.log(
                "Unnormalized macro_avg_acc : Average accuracy across all types (Including mentions without the correct cui in top candidates)",
                self.test_results["unnormalized_macro_avg_acc"],
                sync_dist=True,
            )

        self.test_results = convert_defaultdict(self.test_results)

        # Store results for evaluation object
        eval_file_name = os.path.join(
            self.hparams["output_path"],
            f"crossencoder_output_eval",
            self.hparams["experiment"],
        )
        with open(f"{eval_file_name}.json", "w") as f:
            json.dump(self.test_results, f, indent=2)
            print(f"\nPredictions overview saved at: {eval_file_name}.json")

    def on_train_start(self):
        self.start_time = time.time()

    def on_train_end(self):
        execution_time = (time.time() - self.start_time) / 60
        logging.info(f"The training took {execution_time} minutes")


def main(args):
    print("Current seed:", args["seed"])

    data_module = CrossEncoderDataModule(params=args)

    model = CrossEncoder(params=args)

    current_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    model_checkpoint = ModelCheckpoint(
        monitor="max_acc",  # Metric to monitor
        dirpath=args["output_path"],  # Directory to save the model
        filename=f"{current_time}-{{epoch}}-{{max_acc:.2f}}",  # Saves the model with epoch and val_loss in the filename
        save_top_k=1,  # Number of best models to save; -1 means save all of them
        mode="max",  # 'max' means the highest max_acc will be considered as the best model
        verbose=True,  # Logs a message whenever a model checkpoint is saved
    )

    wandb_logger = WandbLogger(project=args["experiment"])

    trainer = L.Trainer(
        limit_val_batches=1,
        # num_sanity_val_steps=0,
        # fast_dev_run=True,
        max_epochs=args["num_train_epochs"],
        devices=args["devices"],
        accelerator="gpu",
        strategy="ddp_find_unused_parameters_true",
        enable_progress_bar=True,
        callbacks=[model_checkpoint],
        accumulate_grad_batches=args["gradient_accumulation_steps"],
        precision="16-mixed",
        check_val_every_n_epoch=1,
        logger=wandb_logger,
    )

    trainer.fit(model, datamodule=data_module)


if __name__ == "__main__":
    parser = BlinkParser(add_model_args=True)
    parser.add_training_args()
    args = parser.parse_args()
    print(args)
    main(args.__dict__)
